import numpy as np
import mne
from copy import deepcopy
from ..util import find_corner
from .base import BaseSolver, InverseOperator
from time import time

class SolverMVAB(BaseSolver):
    ''' Class for the Minimum Variance Adaptive Beamformer (MVAB) inverse
        solution [1].
    
    Attributes
    ----------

    References
    ----------
    [1] Vorobyov, S. A. (2013). Principles of minimum variance robust adaptive
    beamforming design. Signal Processing, 93(12), 3264-3277.

    '''
    def __init__(self, name="Minimum Variance Adaptive Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience

        '''

        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)
        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = self.leadfield.shape

        y = data
        C = y@y.T
        I = np.identity(n_chans)

        self.get_alphas(reference=C)
        
        inverse_operators = []
        for alpha in self.alphas:
            R_inv = np.linalg.inv(C + alpha * I)
            inverse_operator = 1/(leadfield.T @ R_inv @ leadfield + alpha * np.identity(n_dipoles)) @ leadfield.T @ R_inv

            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverLCMV(BaseSolver):
    ''' Class for the Linearly Constrained Minimum Variance Beamformer (LCMV)
        inverse solution [1].
    
    Attributes
    ----------

    References
    ----------
    [1] Van Veen, B. D., & Buckley, K. M. (1988). Beamforming: A versatile
    approach to spatial filtering. IEEE assp magazine, 5(2), 4-24.
    
    '''
    def __init__(self, name="LCMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha='auto', weight_norm=True, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience

        '''
        self.weight_norm = weight_norm
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)
        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = self.leadfield.shape
        
        y = data
        
        I = np.identity(n_chans)
        C = y@y.T
        
        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        self.alphas = np.logspace(-4, 1, self.n_reg_params) * np.diagonal(y@y.T).mean()

        inverse_operators = []
        for alpha in self.alphas:
            
            C_inv = np.linalg.inv(C + alpha*I)

            # W = (C_inv @ leadfield) / np.diagonal(leadfield.T @ C_inv @ leadfield)
            upper = C_inv @ leadfield
            lower = np.einsum('ij,jk,ki->i', leadfield.T, C_inv, leadfield)
            W = upper / lower
            
            # C_inv_L = C_inv @ leadfield
            # diagonal_elements = np.einsum('ij,ji->i', leadfield.T, C_inv_L)
            # W = C_inv_L / diagonal_elements
            
            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)
            
            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self


class SolverSMV(BaseSolver):
    ''' Class for the Standardized Minimum Variance (SMV) Beamformer inverse
        solution [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones, R.
    (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.

    '''
    def __init__(self, name="SMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience

        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = self.leadfield.shape

        self.weight_norm = weight_norm

        y = data
        I = np.identity(n_chans)
        
        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y@y.T
        self.alphas = self.get_alphas(reference=C)

        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            W = (C_inv @ leadfield) / np.sqrt(np.diagonal(leadfield.T @ C_inv @ leadfield))
            
            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)
            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverWNMV(BaseSolver):
    ''' Class for the Weight-normalized Minimum Variance (WNMV) Beamformer
        inverse solution [1].
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones,
    R. (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.

    '''
    def __init__(self, name="WNMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = self.leadfield.shape

        self.weight_norm = weight_norm
        y = data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y@y.T
        self.alphas = self.get_alphas(reference=C)
        
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            C_inv_2 = np.linalg.inv(C_inv)
            W = (C_inv @ leadfield) / np.sqrt(abs(np.diagonal(leadfield.T @ C_inv_2 @ leadfield)))

            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverHOCMV(BaseSolver):
    ''' Class for the Higher-Order Covariance Minimum Variance (HOCMV)
        Beamformer inverse solution [1].
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones,
    R. (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.

    '''
    def __init__(self, name="HOCMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, alpha='auto', order=3, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        order : int
            The order of the covariance matrix. Should be a positive integer not
            evenly divisible by two {3, 5, 7, ...}
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = self.leadfield.shape
        
        self.weight_norm = weight_norm
        
        y = data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y@y.T
        self.alphas = self.get_alphas(reference=C)
        
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            C_inv_n = deepcopy(C_inv)
            for _ in range(order-1):
                C_inv_n = np.linalg.inv(C_inv_n)

            W = (C_inv @ leadfield) / np.sqrt(abs(np.diagonal(leadfield.T @ C_inv_n @ leadfield)))
            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverESMV(BaseSolver):
    ''' Class for the Eigenspace-based Minimum Variance (ESMV) Beamformer
        inverse solution [1].
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones,
    R. (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.
    
    '''
    def __init__(self, name="ESMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = leadfield.shape
        

        y = data
        I = np.identity(n_chans)
        
        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y@y.T
        self.alphas = self.get_alphas(reference=C)
        subspace = self.reduce_rank_matrix(C)

        
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            C_inv_leadfield = C_inv @ leadfield
            diag_elements = np.einsum('ij,ji->i', leadfield.T, C_inv_leadfield)
            W_mv = C_inv_leadfield / diag_elements
            W = subspace @ W_mv

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

    def reduce_rank_matrix(self, C):
        # Calculte eigenvectors and eigenvalues
        U, s, _ = np.linalg.svd(C, full_matrices=False)
        
        # Find optimal rank using L-curve Corner approach:
        n_comp_l = find_corner(np.arange(len(s)), s)
        
        # Find optimal rank using drop-off approach:
        s_ = s / s.max()
        n_comp_d = np.where( abs(np.diff(s_)) < 0.001 )[0]
        if len(n_comp_d) > 0:
            n_comp_d = n_comp_d[0] + 2
        else:
            n_comp_d = n_comp_l
        
        # Kaiser Rule
        n_comp_k = np.where(s < np.mean(s))[0][0]
        print(n_comp_k, n_comp_l, n_comp_d)
        
        # Combine the approaches
        n_comp = np.ceil((n_comp_d + n_comp_l + n_comp_k)/3).astype(int)
        
        # Transform data
        subspace = U[:, :n_comp] @ U[:, :n_comp].T
        return subspace

class SolverMCMV(BaseSolver):
    ''' Class for the Multiple Constrained Minimum Variance (MCMV) Beamformer
    inverse solution [1].
    
    Attributes
    ----------

    
    References
    ----------
    [1] Nunes, A. S., Moiseev, A., Kozhemiako, N., Cheung, T., Ribary, U., &
    Doesburg, S. M. (2020). Multiple constrained minimum variance beamformer
    (MCMV) performance in connectivity analyses. NeuroImage, 208, 116386.

    '''
    def __init__(self, name="MCMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, noise_cov=None, alpha='auto', verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        # leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = leadfield.shape

        if noise_cov is None:
            noise_cov = np.identity(n_chans)
            
        self.weight_norm = weight_norm

        y = data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y @ y.T
        self.alphas = self.get_alphas(reference=C)
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            leadfield_C_inv = leadfield.T @ C_inv

            # Use np.einsum to compute the diagonal elements
            diag_elements = np.einsum('ij,ji->i', leadfield_C_inv, leadfield)

            W = C_inv @ leadfield * (1.0 / diag_elements)
            # W = C_inv @ leadfield @ np.linalg.pinv(leadfield.T @ C_inv @ leadfield)

            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]

        return self
    
class SolverUnitNoiseGain(BaseSolver):
    ''' Class for the Unit Noise Gain (UNIG) Beamformer
    inverse solution [1].
    
    Attributes
    ----------

    
    References
    ----------
    [1] 
    '''
    def __init__(self, name="UNIG Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, noise_cov=None, alpha='auto', verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        # leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = leadfield.shape

        if noise_cov is None:
            noise_cov = np.identity(n_chans)
            
        self.weight_norm = weight_norm

        y = data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y @ y.T
        self.alphas = self.get_alphas(reference=C)
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            C_inv_inv = np.linalg.inv(C_inv)
            leadfield_C_inv_inv = leadfield.T @ C_inv_inv

            # Use np.einsum to compute the diagonal elements
            diag_elements = np.einsum('ij,ji->i', leadfield_C_inv_inv, leadfield)

            # W = C_inv @ leadfield * (1.0 / diag_elements)
            W = C_inv @ leadfield * (1 / np.sqrt(diag_elements))

            # W = C_inv @ leadfield @ np.linalg.pinv(leadfield.T @ C_inv @ leadfield)

            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]

        return self

class SolverHOCMCMV(BaseSolver):
    ''' Class for the Higher-Order Covariance Multiple Constrained Minimum Variance (HOCMCMV)
        Beamformer inverse solution [1].
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Jonmohamadi, Y., Poudel, G., Innes, C., Weiss, D., Krueger, R., & Jones,
    R. (2014). Comparison of beamformers for EEG source signal reconstruction.
    Biomedical Signal Processing and Control, 14, 175-188.

    '''
    def __init__(self, name="HOCMCMV Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, alpha='auto', order=3, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        order : int
            The order of the covariance matrix. Should be a positive integer not
            evenly divisible by two {3, 5, 7, ...}
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = self.leadfield.shape
        
        self.weight_norm = weight_norm
        
        y = data
        I = np.identity(n_chans)

        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        C = y@y.T
        self.alphas = self.get_alphas(reference=C)
        
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            C_inv_n = deepcopy(C_inv)
            for _ in range(order-1):
                C_inv_n = np.linalg.inv(C_inv_n)
            W = C_inv @ leadfield * np.diagonal(np.linalg.inv(leadfield.T @ C_inv_n @ leadfield))
            
            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self
    

class SolverReciPSIICOS(BaseSolver):
    ''' Class for the Reciprocal Phase Shift Invariant Imaging of Coherent
    Sources (ReciPSIICOS) Beamformer inverse solution [1].
    
    Attributes
    ----------


    References
    ----------    
    [1] Kuznetsova, A., Nurislamova, Y., & Ossadtchi, A. (2021). Modified
    covariance beamformer for solving MEG inverse problem in the environment
    with correlated sources. Neuroimage, 228, 117677.

    '''
    def __init__(self, name="ReciPSIICOS", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, K=3, alpha='auto', verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience

        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        leadfield = self.leadfield
        leadfield /= np.linalg.norm(leadfield, axis=0)
        n_chans, n_dipoles = leadfield.shape
        
        self.weight_norm = weight_norm
        
        y = data
        C = y@y.T

        # Step 1
        G_pwpr = self.construct_G_pwpr(leadfield.T)
        
        # Step 2
        P = self.compute_projector(G_pwpr, K)

        # Step 3
        C_projected = self.apply_projection(P, C)

        # Step 4: SVD of C_projected and reconstruct with absolute eigenvalues
        E, A, _ = np.linalg.svd(C_projected, full_matrices=False)
        C = E @ np.diag(np.abs(A)) @ E.T
        
        I = np.identity(n_chans)
        # Recompute regularization based on the max eigenvalue of the Covariance
        # Matrix (opposed to that of the leadfield)
        self.alphas = self.get_alphas(reference=C)

        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(C + alpha * I)
            leadfield_C_inv = leadfield.T @ C_inv

            # MCMV
            diag_elements = np.einsum('ij,ji->i', leadfield_C_inv, leadfield)
            W = C_inv @ leadfield * (1.0 / diag_elements)

            #LCMV
            # upper = C_inv @ leadfield
            # lower = np.einsum('ij,jk,ki->i', leadfield.T, C_inv, leadfield)
            # W = upper / lower

            if self.weight_norm:
                W /= np.linalg.norm(W, axis=0)

            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self
    
    @staticmethod
    def construct_G_pwpr(g_list):
        # g_list is a list of topography vectors g_i for N sources
        N = len(g_list)
        q_list = [np.outer(g, g).flatten() for g in g_list]
        G_pwpr = np.column_stack(q_list)
        return G_pwpr

    @staticmethod
    def compute_projector(G_pwpr, K):
        U, S, Vt = np.linalg.svd(G_pwpr, full_matrices=False)
        U_K = U[:, :K]  # Keep only the first K singular vectors
        P = U_K @ U_K.T  # Projector matrix
        return P

    @staticmethod
    def apply_projection(P, C_x):
        # Vectorize the sensor-space covariance matrix C_x
        vec_Cx = C_x.flatten()
        # Project the vectorized covariance matrix
        projected_vec_Cx = P @ vec_Cx
        # Reshape back to matrix form
        C_x_projected = np.reshape(projected_vec_Cx, C_x.shape)
        return C_x_projected


class SolverSAM(BaseSolver):
    ''' Class for the Synthetic Aperture Magnetometry Beamformer (SAM) inverse
    solution [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Robinson, S. E. V. J. (1999). Functional neuroimaging by synthetic
    aperture magnetometry (SAM). Recent advances in biomagnetism.

    '''
    def __init__(self, name="SAM Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, alpha='auto', verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        weight_norm : bool
            Normalize the filter weight matrix W to unit length of the columns.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)

        self.weight_norm = weight_norm
        leadfield = self.leadfield
        n_chans, n_dipoles = leadfield.shape
        

        y = data
        I = np.identity(n_chans)
  
        
        inverse_operators = []
        for alpha in self.alphas:
            C_inv = np.linalg.inv(y@y.T + alpha * I)
            W = []
            for i in range(n_dipoles):
                l = leadfield[:, i][:, np.newaxis]
                w = (C_inv@l) / (l.T@C_inv@l)
                W.append(w)
            W = np.stack(W, axis=1)[:, :, 0]
            if self.weight_norm:
                W = W / np.linalg.norm(W, axis=0)
            inverse_operator = W.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverEBB(BaseSolver):
    """
    Empirical Bayesian Beamformer (EBB) solver for M/EEG inverse problem.
    """
    def __init__(self, name="Empirical Bayesian Beamformer", reduce_rank=True, rank="auto", **kwargs):
        self.name = name
        return super().__init__(reduce_rank=reduce_rank, rank=rank, **kwargs)
        
    def make_inverse_operator(self, forward, mne_obj, *args, weight_norm=True, noise_cov=None, alpha='auto', **kwargs):
        """
        Solve the inverse problem using the Empirical Bayesian Beamformer method.

        Parameters:
        -----------
        data : array, shape (n_channels, n_times)
            The sensor data.
        forward : array, shape (n_channels, n_sources)
            The forward solution.
        noise_cov : array, shape (n_channels, n_channels), optional
            The noise covariance matrix.

        Returns:
        --------
        sources : array, shape (n_sources, n_times)
            The estimated source time series.
        """
        print(type(forward))
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)

        data = self.unpack_data_obj(mne_obj)
        leadfield = self.leadfield
        n_channels, n_times = data.shape
        n_sources = leadfield.shape[1]
        leadfield /= np.linalg.norm(leadfield, axis=0)
        

        # Compute data covariance
        data_cov = np.cov(data)

        # handle noise_cov
        if noise_cov is None:
            noise_cov = np.identity(n_channels)

        # Initialize source covariance
        inverse_operators = []
        self.alphas = self.get_alphas(reference=leadfield@leadfield.T)
        for alpha in self.alphas:
            source_cov = np.eye(n_sources)
            # Iterative process
            for n_iter in range(100):  # You can adjust the number of iterations
                # Compute regularized inverse
                C = leadfield @ source_cov @ leadfield.T
                C_inv = np.linalg.inv(C + alpha * noise_cov)
                
                # Update source covariance
                W = source_cov @ leadfield.T @ C_inv
                new_source_cov = W @ data_cov @ W.T
                if weight_norm:
                    W /= np.linalg.norm(W, axis=0)
                
                # Check convergence
                if np.allclose(new_source_cov, source_cov, rtol=1e-9):
                    print(f"Converged after {n_iter} iterations")
                    break
                
                source_cov = new_source_cov

            # Compute final beamformer weights
            W = source_cov @ leadfield.T @ np.linalg.inv(leadfield @ source_cov @ leadfield.T + noise_cov)
            inverse_operators.append(W)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]

        return self


import numpy as np
import mne
from copy import deepcopy
from .base import BaseSolver, InverseOperator
from scipy.sparse.csgraph import laplacian

class SolverMNE(BaseSolver):
    ''' Class for the Minimum Norm Estimate (MNE) inverse solution [1]. The
        formulas provided by [2] were used for implementation.

    Attributes
    ----------
    
    References
    ----------
    [1] Pascual-Marqui, R. D. (1999). Review of methods for solving the EEG
    inverse problem. International journal of bioelectromagnetism, 1(1), 75-86.
    
    [2] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
    Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
    problem in EEG source analysis. Journal of neuroengineering and
    rehabilitation, 5(1), 1-33.

    '''
    def __init__(self, name="Minimum Norm Estimate", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha="auto", verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, reference=None, alpha=alpha, **kwargs)
        
        leadfield = self.leadfield
        n_chans, _ = leadfield.shape
        
        LLT = leadfield @ leadfield.T
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = leadfield.T @ np.linalg.inv(LLT + alpha * np.identity(n_chans))
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverGFTMNE(BaseSolver):
    ''' Class for the Minimum Norm Estimate (MNE) inverse solution [1]. The
        formulas provided by [2] were used for implementation.

    Attributes
    ----------
    
    References
    ----------
    [1] Pascual-Marqui, R. D. (1999). Review of methods for solving the EEG
    inverse problem. International journal of bioelectromagnetism, 1(1), 75-86.
    
    [2] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
    Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
    problem in EEG source analysis. Journal of neuroengineering and
    rehabilitation, 5(1), 1-33.

    '''
    def __init__(self, name="Minimum Norm Estimate", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha="auto", verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, reference=None, alpha=alpha, **kwargs)
        
        leadfield = self.leadfield
        n_chans, _ = leadfield.shape

        # Get Adjacency matrix
        
        adjacency = mne.spatial_src_adjacency(forward['src'], verbose=0)
        lap = laplacian(adjacency)
        
        U, eigenvalues , _ = np.linalg.svd(lap.toarray(), full_matrices=False)
        # Filter
        num_eigenvalues = len(eigenvalues)
        cutoff_index = 100  # int(num_eigenvalues * 0.3)  # 30% cutoff
        # U = U[:, :cutoff_index]
        U = np.real(U[:, -cutoff_index:])

        # Transform leadfield
        leadfield_gft = leadfield @ U

        LLT = leadfield_gft @ leadfield_gft.T
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = leadfield_gft.T @ np.linalg.inv(LLT + alpha * np.identity(n_chans))
            inverse_operators.append(U @ inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverWMNE(BaseSolver):
    ''' Class for the Weighted Minimum Norm Estimate (wMNE) inverse solution
        [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
    Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
    problem in EEG source analysis. Journal of neuroengineering and
    rehabilitation, 5(1), 1-33.
    '''
    def __init__(self, name="Weighted Minimum Norm Estimate", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        W = np.diag(np.linalg.norm(self.leadfield, axis=0))
        WTW = np.linalg.inv(W.T @ W)
        LWTWL = self.leadfield @ WTW @ self.leadfield.T
        n_chans, _ = self.leadfield.shape

      
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = WTW @ self.leadfield.T  @ np.linalg.inv(LWTWL + alpha * np.identity(n_chans))
            inverse_operators.append(inverse_operator)
        
        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverDSPM(BaseSolver):
    ''' Class for the Dynamic Statistical Parametric Mapping (dSPM) inverse
        solution [1,2].  The formulas provided by [3] were used for
        implementation.
    
    Attributes
    ----------
    
    References
    ----------
    [1] Dale, A. M., Liu, A. K., Fischl, B. R., Buckner, R. L., Belliveau, J.
    W., Lewine, J. D., & Halgren, E. (2000). Dynamic statistical parametric
    mapping: combining fMRI and MEG for high-resolution imaging of cortical
    activity. neuron, 26(1), 55-67.

    [2] Dale, A. M., Fischl, B., & Sereno, M. I. (1999). Cortical surface-based
    analysis: I. Segmentation and surface reconstruction. Neuroimage, 9(2),
    179-194.

    [3] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
    Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
    problem in EEG source analysis. Journal of neuroengineering and
    rehabilitation, 5(1), 1-33.
    '''
    def __init__(self, name="Dynamic Statistical Parametric Mapping", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha=0.01, noise_cov=None, source_cov=None,
                            verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        noise_cov : numpy.ndarray
            The noise covariance matrix (channels x channels).
        source_cov : numpy.ndarray
            The source covariance matrix (dipoles x dipoles). This can be used if
            prior information, e.g., from fMRI images, is available.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        n_chans, n_dipoles = self.leadfield.shape

        if noise_cov is None:
            noise_cov = np.identity(n_chans)
        if source_cov is None:
            source_cov = np.identity(n_dipoles)
        

        inverse_operators = []
        leadfield_source_cov = source_cov @ self.leadfield.T
        LLS = self.leadfield @ leadfield_source_cov
        for alpha in self.alphas:
            K = leadfield_source_cov @ np.linalg.inv(LLS + alpha * noise_cov)
            W_dSPM = np.diag( np.sqrt( 1 / np.diagonal(K @ noise_cov @ K.T) ) )
            inverse_operator = W_dSPM @ K
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self
  
class SolverMinimumL1Norm(BaseSolver):
    ''' Class for the Minimum Current Estimate (MCE) inverse solution using the
        FISTA solver [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Beck, A., & Teboulle, M. (2009). A fast iterative shrinkage-thresholding
    algorithm for linear inverse problems. SIAM journal on imaging sciences,
    2(1), 183-202.
    '''
    def __init__(self, name="Minimum Current Estimate", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', max_iter=1000, noise_cov=None, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        n_chans = self.leadfield.shape[0]
        if noise_cov is None:
            noise_cov = np.identity(n_chans)

        self.noise_cov = noise_cov
        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj, max_iter=1000, 
                               l1_reg=1e-3, l2_reg=0, tol=1e-2) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        l1_reg : float
            Controls the spatial L1 regularization 
        l2_reg : float
            Controls the spatial L2 regularization 
        tol : float
            Tolerance at which convergence is met.
        
        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        '''
        data = self.unpack_data_obj(mne_obj)
        source_mat = self.fista_wrap(data, max_iter=max_iter, 
                                    l1_reg=l1_reg, l2_reg=l2_reg, tol=tol)
        stc = self.source_to_object(source_mat)
        return stc
    
    def fista_wrap(self, y_mat, max_iter=1000, l1_reg=1e-3, l2_reg=0, tol=1e-2):
        srcs = []
        for y in y_mat.T:
            srcs.append ( self.fista(y, max_iter=max_iter, 
                                    l1_reg=l1_reg, l2_reg=l2_reg, 
                                    tol=tol) 
                                    )
        return np.stack(srcs, axis=1)


    def fista(self, y, l1_reg=1e-3, l2_reg=0, max_iter=1000, tol=1e-2):
        """
        Solves the EEG inverse problem:
            min_x ||y - Ax||_2^2 + l1_reg * ||x||_1 + l2_reg * ||x||_2^2
        using the FISTA algorithm.
        
        Parameters
        ----------
        y : ndarray, shape (m,)
            EEG measurements.
        A : ndarray, shape (m, n)
            Forward model.
        x0 : ndarray, shape (n,)
            Initial guess for the CSDs.
        l1_reg : float, optional (default: 1e-3)
            L1 regularization strength.
        l2_reg : float, optional (default: 0)
            L2 regularization strength.
        max_iter : int, optional (default: 1000)
            Maximum number of iterations to run.
        tol : float, optional (default: 1e-6)
            Tolerance for the stopping criteria.
        
        Returns
        -------
        x : ndarray, shape (n,)
            Estimated CSDs.
        """
        
        def grad_f(x):
            """Gradient of the objective function"""
            return x - y + l1_reg * np.sign(x) + l2_reg * x

        A = deepcopy(self.leadfield)
        A /= np.linalg.norm(A, axis=0)

        y_scaled = y.copy()
        # Rereference
        y_scaled -= y_scaled.mean()
        # Scale to unit norm
        norm_y = np.linalg.norm(y_scaled)
        y_scaled /= norm_y
        
        # Calculate initial guess
        x0 = np.linalg.pinv(A) @ y_scaled
        # Scale to unit norm
        x0 /= np.linalg.norm(x0)

        x = x0.copy()
        y = x0.copy()
        
        t = 1.0
        lr = 1.0
        for i in range(max_iter):
            x_prev = x.copy()
            # Gradient descent step
            x = y - lr * grad_f(y)
            # Soft thresholding step
            x = self.soft_threshold(x, l1_reg * lr)
            # Update y and t
            t_prev = t
            t = (1 + (1 + 4 * t**2)**0.5) / 2
            y = x + (t_prev - 1) / t * (x - x_prev)

            if np.linalg.norm(y) == 0:
                break
            # Check stopping criteria
            if np.linalg.norm(x - x_prev) < tol:
                break
        # Rescale source
        x = x * norm_y

        return x
        

    @staticmethod
    def soft_threshold(x, alpha):
        """
        Applies the soft thresholding operator to x with threshold alpha.
        
        Parameters
        ----------
        x : ndarray, shape (n,)
            Input array.
        alpha : float
            Threshold.
        
        Returns
        -------
        y : ndarray, shape (n,)
            Output array.
        """
        y = np.sign(x) * np.maximum(np.abs(x) - alpha, 0)
        return y


class SolverGFTMinimumL1Norm(BaseSolver):
    ''' Class for the Minimum Current Estimate (MCE) inverse solution using the
        FISTA solver [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Beck, A., & Teboulle, M. (2009). A fast iterative shrinkage-thresholding
    algorithm for linear inverse problems. SIAM journal on imaging sciences,
    2(1), 183-202.
    '''
    def __init__(self, name="Minimum Current Estimate", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', max_iter=1000, noise_cov=None, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        n_chans = self.leadfield.shape[0]
        if noise_cov is None:
            noise_cov = np.identity(n_chans)
        
        adjacency = mne.spatial_src_adjacency(forward['src'], verbose=0)
        lap = laplacian(adjacency)
        # U, eigenvalues , _ = np.linalg.svd(lap.toarray(), full_matrices=False)
        eigenvalues, U = np.linalg.eig(lap.toarray())
        cutoff_index = int(len(eigenvalues) * 0.3)  # 30% cutoff
        U = U[:, -cutoff_index:]
        self.U = np.real(U)

        self.noise_cov = noise_cov
        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj, max_iter=1000, 
                               l1_reg=1e-3, l2_reg=0, tol=1e-2) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        l1_reg : float
            Controls the spatial L1 regularization 
        l2_reg : float
            Controls the spatial L2 regularization 
        tol : float
            Tolerance at which convergence is met.
        
        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        '''
        data = self.unpack_data_obj(mne_obj)
        source_mat = self.fista_wrap(data, max_iter=max_iter, 
                                    l1_reg=l1_reg, l2_reg=l2_reg, tol=tol)
        stc = self.source_to_object(source_mat)
        return stc
    
    def fista_wrap(self, y_mat, max_iter=1000, l1_reg=1e-3, l2_reg=0, tol=1e-2):
        srcs = []
        for y in y_mat.T:
            srcs.append ( self.fista(y, max_iter=max_iter, 
                                    l1_reg=l1_reg, l2_reg=l2_reg, 
                                    tol=tol) 
                                    )
        return np.stack(srcs, axis=1)


    def fista(self, y, l1_reg=1e-3, l2_reg=0, max_iter=1000, tol=1e-2):
        """
        Solves the EEG inverse problem:
            min_x ||y - Ax||_2^2 + l1_reg * ||x||_1 + l2_reg * ||x||_2^2
        using the FISTA algorithm.
        
        Parameters
        ----------
        y : ndarray, shape (m,)
            EEG measurements.
        A : ndarray, shape (m, n)
            Forward model.
        x0 : ndarray, shape (n,)
            Initial guess for the CSDs.
        l1_reg : float, optional (default: 1e-3)
            L1 regularization strength.
        l2_reg : float, optional (default: 0)
            L2 regularization strength.
        max_iter : int, optional (default: 1000)
            Maximum number of iterations to run.
        tol : float, optional (default: 1e-6)
            Tolerance for the stopping criteria.
        
        Returns
        -------
        x : ndarray, shape (n,)
            Estimated CSDs.
        """
        
        def grad_f(x):
            """Gradient of the objective function"""
            return x - y + l1_reg * np.sign(x) + l2_reg * x

        A = deepcopy(self.leadfield) @ self.U
        # A /= np.linalg.norm(A, axis=0)

        y_scaled = y.copy()
        # Rereference
        y_scaled -= y_scaled.mean()
        # Scale to unit norm
        norm_y = np.linalg.norm(y_scaled)
        y_scaled /= norm_y
        
        # Calculate initial guess
        x0 = np.linalg.pinv(A) @ y_scaled
        # Scale to unit norm
        x0 /= np.linalg.norm(x0)

        x = x0.copy()
        y = x0.copy()
        
        t = 1.0
        lr = 1.0
        for i in range(max_iter):
            x_prev = x.copy()
            # Gradient descent step
            x = y - lr * grad_f(y)
            # Soft thresholding step
            x = self.soft_threshold(x, l1_reg * lr)
            # Update y and t
            t_prev = t
            t = (1 + (1 + 4 * t**2)**0.5) / 2
            y = x + (t_prev - 1) / t * (x - x_prev)

            if np.linalg.norm(y) == 0:
                print("norm is zero")
                x = x_prev
                break
            # Check stopping criteria
            # print(np.linalg.norm(x - x_prev))
            if np.linalg.norm(x - x_prev) < tol:
                print("criterion met")
                break
        # Rescale source
        x = x * norm_y

        return self.U @ x
        

    @staticmethod
    def soft_threshold(x, alpha):
        """
        Applies the soft thresholding operator to x with threshold alpha.
        
        Parameters
        ----------
        x : ndarray, shape (n,)
            Input array.
        alpha : float
            Threshold.
        
        Returns
        -------
        y : ndarray, shape (n,)
            Output array.
        """
        y = np.sign(x) * np.maximum(np.abs(x) - alpha, 0)
        return y

class SolverMinimumL1NormGPT(BaseSolver):
    ''' Class for the Minimum Current Estimate inverse solution using
        interesting code from the Chat GPT AI by openai.com (GPT-solver). 
        
        I (Lukas Hecker) prompted the task to write a sparsified eLORETA-type
        inverse solution and this came up with little adjustments required.

        I can't express how weird it is for me, too.
    
    Attributes
    ----------
    
    References
    ----------
    Open AI chat GPT (openai.com)
    
    '''
    def __init__(self, name="GPT Solver", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.inverse_operators = []
        return self

    def apply_inverse_operator(self, mne_obj, max_iter=1000, 
                               l1_reg=1e-3, tol=1e-2) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        l1_reg : float
            Controls the spatial L1 regularization 
        tol : float
            Tolerance at which convergence is met.
        
        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        '''
        data = self.unpack_data_obj(mne_obj)

        source_mat = self.solver_wrap(data, max_iter=max_iter, 
                                    l1_reg=l1_reg, tol=tol)
        stc = self.source_to_object(source_mat)
        return stc
    
    def solver_wrap(self, y_mat, max_iter=1000, l1_reg=1e-3, tol=1e-2):
        srcs = []
        for y in y_mat.T:
            srcs.append ( self.solve(y, max_iter=max_iter, 
                                    l1_reg=l1_reg,
                                    tol=tol) 
                                    )
        return np.stack(srcs, axis=1)


    def solve(self, y, l1_reg=1e-3, max_iter=1000, tol=1e-2):
        """
        Solves the EEG inverse problem:
            min_x ||y - Ax||_2^2 + l1_reg * ||x||_1 + l2_reg * ||x||_2^2
        using the FISTA algorithm.
        
        Parameters
        ----------
        y : ndarray, shape (m,)
            EEG measurements.
        l1_reg : float, optional (default: 1e-3)
            L1 regularization strength.
        max_iter : int, optional (default: 1000)
            Maximum number of iterations to run.
        tol : float, optional (default: 1e-6)
            Tolerance for the stopping criteria.
        
        Returns
        -------
        x : ndarray, shape (n,)
            Estimated CSDs.
        """
        
        
        A = deepcopy(self.leadfield)
        A /= np.linalg.norm(A, axis=0)

        y_scaled = y.copy()
        # Rereference
        y_scaled -= y_scaled.mean()
        # Scale to unit norm
        norm_y = np.linalg.norm(y_scaled)
        y_scaled /= norm_y
        
        # Calculate initial guess
        x0 = np.linalg.pinv(A) @ y_scaled
        # Scale to unit norm
        x0 /= np.linalg.norm(x0)

        x = x0.copy()
        y = x0.copy()
        
        lr = 1.0
        for i in range(max_iter):
            x_prev = x.copy()
            # Gradient descent step
            x = x - lr*l1_reg * self.grad_f(x, A, y_scaled)
            # Soft thresholding step
            x = self.soft_threshold(x, lr*l1_reg)
            
            # Check stopping criteria
            if np.linalg.norm(x) == 0:
                x = x_prev
                break
            if np.linalg.norm(x - x_prev) < tol:
                break
        # Rescale source
        x = x * norm_y

        return x

    @staticmethod        
    def grad_f(x, A, y_scaled):
        """Gradient of the objective function"""
        return A.T.dot(A.dot(x) - y_scaled) + np.sign(x)

    @staticmethod
    def soft_threshold(x, alpha):
        """
        Applies the soft thresholding operator to x with threshold alpha.
        
        Parameters
        ----------
        x : ndarray, shape (n,)
            Input array.
        alpha : float
            Threshold.
        
        Returns
        -------
        y : ndarray, shape (n,)
            Output array.
        """
        y = np.sign(x) * np.maximum(np.abs(x) - alpha, 0)
        return y

class SolverMinimumL1L2Norm(BaseSolver):
    ''' Class for the Minimum L1-L2 Norm solution (MCE) inverse solution. It
        imposes a L1 norm on the source and L2 on the source time courses.
    
    References
    ----------
    [!] Missing reference - please contact developers if you have it!

    '''
    def __init__(self, name="Minimum L1-L2 Norm", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha=0.01, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        
        return self

    def apply_inverse_operator(self, mne_obj, alpha="auto", 
                               max_iter=100, l1_spatial=1e-3, 
                               l2_temporal=1e-3, tol=1e-6, ) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        l1_spatial : float
            Controls the spatial L1 regularization 
        l2_temporal : float
            Controls the temporal L2 regularization 
        tol : float
            Tolerance at which convergence is met.
        
        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object.
        '''
        
        data = self.unpack_data_obj(mne_obj)

        source_mat = self.fista_eeg(data, alpha=alpha, max_iter=max_iter, 
                                    tol=tol, l1_spatial=l1_spatial, 
                                    l2_temporal=l2_temporal)
        stc = self.source_to_object(source_mat)
        return stc
    
    def fista_eeg(self, y, alpha="auto", l1_spatial=1e-3, l2_temporal=1e-3, 
                  max_iter=1000, tol=1e-6):
        """
        Solves the EEG inverse problem using FISTA with L1 regularization on the spatial
        dimension and L2 regularization on the temporal dimension.
        
        Parameters:
        - A: array of shape (n_sensors, n_sources)
        - y: array of shape (n_sensors, n_timepoints)
        - l1_spatial: float, strength of L1 regularization on the spatial dimension
        - l2_temporal: float, strength of L2 regularization on the temporal dimension
        - max_iter: int, maximum number of iterations
        - tol: float, tolerance for convergence
        
        Returns:
        - x: array of shape (n_sources, n_timepoints), the solution to the EEG inverse problem
        """
        A = self.leadfield
        A -= A.mean(axis=0)
        A /= np.linalg.norm(A, axis=0)

        norm_y = np.linalg.norm(y)
        y -= y.mean(axis=0)
        y_scaled = y / norm_y

        # Regularization
        if alpha == "auto":
            alpha = l1_spatial

        # Initialize x and z to be the same, and set t to 1
        W = np.diag(np.linalg.norm(A, axis=0))
        WTW = np.linalg.inv(W.T @ W)
        LWTWL = A @ WTW @ A.T
        inverse_operator = WTW @ A.T  @ np.linalg.inv(LWTWL + alpha * np.identity(A.shape[0]))
        x = z = inverse_operator @ y_scaled
        
        # x = z = np.linalg.pinv(A) @ y_scaled
        
        x /= np.linalg.norm(x)

        t = 1
        
        # Compute the Lipschitz constant
        L = np.linalg.norm(A, ord=2) ** 2
        
        for i in range(max_iter):
            # Compute the gradient of the smooth part of the objective
            grad = A.T @ (A @ x - y_scaled)
            
            # Compute the proximal operator of the L1 regularization
            x_new = np.sign(x - grad / L) * np.maximum(np.abs(x - grad / L) - l1_spatial / L, 0)
            
            # Compute the proximal operator of the L2 regularization
            x_new = x_new - np.mean(x_new, axis=0)
            x_new = x_new / np.linalg.norm(x_new, ord=2, axis=0) * np.maximum(np.linalg.norm(x_new, ord=2, axis=0) - l2_temporal / L, 0)
            x_new = x_new + np.mean(x_new, axis=0)
            
            # Update t and z
            t_new = (1 + np.sqrt(1 + 4 * t ** 2)) / 2
            z_new = x_new + (t - 1) / t_new * (x_new - x)
            
            # Check for convergence
            if np.linalg.norm(x_new - x) < tol:
                # print("convergence after ", i)
                break
                
            # Update x, t, and z
            x = x_new
            t = t_new
            z = z_new
        # Rescale Sources
        x = x * norm_y

        return x


    @staticmethod
    def calc_norm(x, n_time):
        return np.sqrt( (x**2).sum() / n_time )


from copy import deepcopy
import numpy as np
import mne
import matplotlib.pyplot as plt
import os
import pickle as pkl
import tensorflow as tf
from mne.io.constants import FIFF
from ..util import find_corner

class InverseOperator:
    ''' This class holds the inverse operator, which may be a simple
    numpy.ndarray matrix or some object like an esinet.net()

    Parameters
    ----------
    inverse operator : 
    Return
    ------
    '''

    def __init__(self, inverse_operator, solver_name):
        self.solver_name = solver_name
        self.data = inverse_operator
        self.handle_inverse_operator()
        self.has_multiple_operators()

    def has_multiple_operators(self):
        ''' Check if there are multiple inverse_operators.'''
        if type(self.data) == list:
            if len(self.data) > 1:
                return True
        return False

    def handle_inverse_operator(self):
        if type(self.data) != list:
            self.data = [self.data,]
        self.type = type(self.data[0])
    
    def apply(self, M):
        ''' Apply the precomputed inverse operator to the data matrix M.
        Parameters
        ----------
        M : numpy.ndarray
            The M/EEG data matrix (n_channels, n_timepoints)

        Return
        ------
        J : numpy.ndarray
            The source estimate matrix (n_sources, n_timepoints)
        '''
 
        J = self.data @ M
        if len(J.shape) > 2:
            J = np.squeeze(J)
        return J
        
class BaseSolver:
    '''
    Parameters
    ----------
    regularisation_method : str
        Can be either 
            "GCV"       -> generalized cross validation
            "L"         -> L-Curve method using triangle method
            "Product"   -> Minimal product method
    n_reg_params : int
        The number of regularisation parameters to try. The higher, the 
        more accurate the regularisation and the slower the computations.
    prep_leadfield : bool
        If True -> Apply common average referencing and normalisation of the leadfield columns.
    reduce_rank : bool
        Whether to reduce the rank of the M/EEG data
    rank : str/int
        Can be either int -> select only the <rank> largest eigenvectors of the data
        "auto" -> automatically select the optimal rank using the L-curve method and 
                  an eigenvalue drop-off criterion
    plot_reg : bool
        Plot the regularization parameters.

    '''
    def __init__(self, regularisation_method="GCV", n_reg_params=11, 
        prep_leadfield=False, use_last_alpha=False, rank="auto",
        reduce_rank=False, plot_reg=False, common_average_reference=False, verbose=0):
        self.verbose = verbose

        self.r_values = np.insert(np.logspace(-9, 1, n_reg_params), 0, 0)


        # self.alphas = deepcopy(self.r_values)
        self.n_reg_params = n_reg_params
        self.regularisation_method = regularisation_method
        self.prep_leadfield = prep_leadfield
        self.use_last_alpha = use_last_alpha
        self.last_reg_idx = None
        self.rank = rank
        self.reduce_rank = reduce_rank
        self.plot_reg = plot_reg
        self.made_inverse_operator = False
        self.common_average_reference = common_average_reference

    def make_inverse_operator(self, forward: mne.Forward, *args, alpha="auto", reference=None, **kwargs):
        """ Base function to create the inverse operator based on the forward
            model.

        Parameters
        ----------
        forward : mne.Forward
            The mne Forward model object.
        alpha : ["auto", float]
            If "auto": Try out multiple regularization parameters and choose the
            optimal one, otherwise use the float.


        Return
        ------
        None
        
        """
        self.forward = deepcopy(forward)
        self.prepare_forward()
        self.alpha = alpha
        self.alphas = self.get_alphas(reference=reference)
        self.made_inverse_operator = True

    def store_obj_information(self, mne_obj):
        if hasattr(mne_obj, "tmin"):
            self.tmin = mne_obj.tmin
        else:
            self.tmin = 0
        
        self.obj_info = mne_obj.info
        
    def apply_inverse_operator(self, mne_obj) -> mne.SourceEstimate:
        ''' Apply the inverse operator
        
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        
        Return
        ------
        stc : mne.SourceEstimate
            The mne SourceEstimate object.
        
        '''
        
        data = self.unpack_data_obj(mne_obj)

        # if type(self.inverse_operators[0].data) == list:
        #     M = data
            
        #     maximum_a_posteriori, A, S = self.inverse_operators[0].data
        #     # transform data M with spatial (A) and temporal (S) projector
        #     M_ = A @ M @ S
        #     # invert transformed data M_ to tansformed sources J_
        #     J_ = maximum_a_posteriori @ M_
        #     # Project tansformed sources J_ back to original time frame using temporal projector S
        #     source_mat = J_ @ S.T 
        #     stc = self.source_to_object(source_mat)
        #     return stc
        
        if self.use_last_alpha and self.last_reg_idx is not None:
            source_mat = self.inverse_operators[self.last_reg_idx].apply( data ) 
            
        else:
            if self.regularisation_method.lower() == "l":
                source_mat, idx = self.regularise_lcurve(data, plot=self.plot_reg)
                self.last_reg_idx = idx
            elif self.regularisation_method.lower() == "gcv":
                source_mat, idx = self.regularise_gcv(data, plot=self.plot_reg)
                self.last_reg_idx = idx
            elif self.regularisation_method.lower() == "product":
                source_mat, idx = self.regularise_product(data, plot=self.plot_reg)
                self.last_reg_idx = idx
            else:
                msg = f"{self.regularisation_method} is no valid regularisation method."
                raise AttributeError(msg)
            
        stc = self.source_to_object(source_mat)
        return stc
        
    
    def prep_data(self, mne_obj):
        if not mne_obj.proj and "eeg" in mne_obj.get_channel_types() and self.common_average_reference:
            mne_obj.set_eeg_reference("average", projection=True, verbose=0).apply_proj(verbose=0)
        
        return mne_obj

    def unpack_data_obj(self, mne_obj, pick_types=None):
        ''' Unpacks the mne data object and returns the data.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.EvokedArray, mne.Epochs, mne.EpochsArray, mne.Raw]

        Return
        ------
        data : numpy.ndarray
            The M/EEG data matrix.

        '''

        type_list = [mne.Evoked, mne.EvokedArray, mne.Epochs, mne.EpochsArray, mne.io.Raw, mne.io.RawArray,]  # mne.io.brainvision.brainvision.RawBrainVision]
        if pick_types is None:
            pick_types = ["meg", "eeg", "fnirs"]
        assert type(pick_types) != dict(), f"pick_types must be of type str or list(str), but is of type dict()"
        
        # Prepare Data
        mne_obj = self.prep_data(mne_obj)
        mne_obj_meeg = mne_obj.copy().pick(pick_types)

        channels_in_fwd = self.forward.ch_names
        channels_in_mne_obj = mne_obj_meeg.ch_names
        picks = self.select_list_intersection(channels_in_fwd, channels_in_mne_obj)
        
        # Select only data channels in mne_obj
        mne_obj_meeg.pick(picks)
        
        # Store original forward model for later
        self.forward_original = deepcopy(self.forward)

        # Select only available data channels in forward
        self.forward = self.forward.pick_channels(picks)
        
        # Prepare the potentially new forward model
        self.prepare_forward()

        # Test if ch_names in forward model and mne_obj_meeg are equal
        assert self.forward.ch_names == mne_obj_meeg.ch_names, "channels available in mne object are not equal to those present in the forward model."
        assert len(self.forward.ch_names) > 1, "forward model contains only a single channel"

        # check if the object is an evoked object
        if isinstance(mne_obj, (mne.Evoked, mne.EvokedArray)):
            # handle evoked object
            data = mne_obj_meeg.data
        
        # check if the object is a raw object
        elif isinstance(mne_obj, (mne.Epochs, mne.EpochsArray)):
            data = mne_obj_meeg.average().data
        
        # check if the object is a raw object
        elif isinstance(mne_obj, (mne.io.Raw, mne.io.RawArray, mne.io.brainvision.brainvision.RawBrainVision)):
            # handle raw object
            data = mne_obj_meeg._data
            # data = mne_obj_meeg.get_data()

        # handle other cases
        else:
            msg = f"mne_obj is of type {type(mne_obj)} but needs to be one of the following types: {type_list}"
            raise AttributeError(msg)
        
        self.store_obj_information(mne_obj)

        if self.reduce_rank:
            data = self.select_signal_subspace(data, rank=self.rank)
        
        return data
    
    @staticmethod
    def select_list_intersection(list1, list2):
        new_list = []
        for element in list1:
            if element in list2:
                new_list.append(element)
        return new_list

    def get_alphas(self, reference=None):
        ''' Create list of regularization parameters (alphas) based on the
        largest eigenvalue of the leadfield or some reference matrix.

        Parameters
        ----------
        reference : [None, numpy.ndarray]
            If None: use leadfield to calculate regularization parameters, else
            use reference matrix (e.g., M/EEG covariance matrix).
        
        Return
        ------
        alphas : list
            List of regularization parameters (alphas)

        '''
        if reference is None:
            L = self.leadfield
            _, eigs, _ = np.linalg.svd(L@L.T, full_matrices=False) 
        else:
            _, eigs, _ = np.linalg.svd(reference, full_matrices=False)
        self.max_eig = eigs.max()

        if self.alpha == "auto":
            alphas = list(self.max_eig * self.r_values)
        else:
            alphas = [self.alpha*self.max_eig, ]
        return alphas

    def regularise_lcurve(self, M, plot=False):
        """ Find optimally regularized inverse solution using the L-Curve method [1].
        
        Parameters
        ----------
        M : numpy.ndarray
            The M/EEG data matrix (n_channels, n_timepoints)
        
        Return
        ------
        source_mat : numpy.ndarray
            The inverse solution  (dipoles x time points)
        optimum_idx : int
            The index of the selected (optimal) regularization parameter
        
        References
        ----------
        [1] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
        Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
        problem in EEG source analysis. Journal of neuroengineering and
        rehabilitation, 5(1), 1-33.
        
        """

        leadfield = self.leadfield
        source_mats = [inverse_operator.apply( M ) for inverse_operator in self.inverse_operators]
        
        # M -= M.mean(axis=0)
        # leadfield -= leadfield.mean(axis=0)
        

        # l2_norms = [np.log(np.linalg.norm( leadfield @ source_mat )) for source_mat in source_mats]
        # l2_norms = [np.log(np.linalg.norm(source_mat )) for source_mat in source_mats]
        l2_norms = [np.linalg.norm( source_mat ) for source_mat in source_mats]
        
        
        # residual_norms = [np.log(np.linalg.norm( leadfield @ source_mat - M )) for source_mat in source_mats]
        residual_norms = [np.linalg.norm( leadfield @ source_mat - M ) for source_mat in source_mats]



        # Filter non-monotonic decreasing values
        # bad_idc = self.filter_norms(self.r_values, l2_norms)
        # l2_norms = np.delete(l2_norms, bad_idc)
        # source_mats = self.delete_from_list(source_mats, bad_idc)
        
        optimum_idx = self.find_corner(l2_norms, residual_norms)
        
        # curvature = self.get_curvature(residual_norms, l2_norms)
        # print(curvature)

        # optimum_idx = np.argmax(curvature)


        source_mat = source_mats[optimum_idx]
        if plot:
            plt.figure()
            plt.plot(residual_norms, l2_norms, 'ok')
            plt.plot(residual_norms[optimum_idx], l2_norms[optimum_idx], 'r*')
            alpha = self.alphas[optimum_idx]
            plt.title(f"L-Curve: {alpha}")

        return source_mat, optimum_idx
        
    @staticmethod
    def get_curvature(x, y):
        
        x_t = np.gradient(x)
        y_t = np.gradient(y)
        vel = np.array([ [x_t[i], y_t[i]] for i in range(x_t.size)])
        speed = np.sqrt(x_t * x_t + y_t * y_t)
        tangent = np.array([1/speed] * 2).transpose() * vel

        ss_t = np.gradient(speed)
        xx_t = np.gradient(x_t)
        yy_t = np.gradient(y_t)

        curvature_val = np.abs(xx_t * y_t - x_t * yy_t) / (x_t * x_t + y_t * y_t)**1.5

        return curvature_val

    def regularise_gcv(self, M, plot=False):
        """ Find optimally regularized inverse solution using the generalized
        cross-validation method [1].
        
        Parameters
        ----------
        M : numpy.ndarray
            The M/EEG data matrix (n_channels, n_timepoints)
        
        Return
        ------
        source_mat : numpy.ndarray
            The inverse solution  (dipoles x time points)
        optimum_idx : int
            The index of the selected (optimal) regularization parameter
        
        References
        ----------
        [1] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
        Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
        problem in EEG source analysis. Journal of neuroengineering and
        rehabilitation, 5(1), 1-33.

        """
        n_chans = self.leadfield.shape[0]
        # Common Average Reference
        # M -= M.mean(axis=0)
        
        I = np.identity(n_chans)
        gcv_values = []
        for inverse_operator in self.inverse_operators:
            x = inverse_operator.apply( M )
            M_hat = self.leadfield @ x 
            # M_hat -= M_hat.mean(axis=0)
            residual_norm = np.linalg.norm(M_hat- M)
            denom = np.trace(I - self.leadfield @ inverse_operator.data[0])**2
    
            gcv_value = residual_norm / denom
            gcv_values.append(gcv_value)
            # print(np.linalg.norm(x), gcv_value)
        # Filter gcv_values that first increase
        keep_idx = 0
        # if len(np.where((np.diff(gcv_values)<0))[0]) == 0:
        #     if not np.isnan(gcv_values[0]):
        #         keep_idx = 0
        #     else:
        #         keep_idx = 1
        # else:
        #     keep_idx = np.where((np.diff(gcv_values)<0))[0][0]
            
        optimum_idx = np.argmin(gcv_values[keep_idx:])+keep_idx

        # optimum_idx = np.argmin(gcv_values[1:])+1
        if plot:
            plt.figure()
            plt.plot(self.alphas, gcv_values, 'ok')
            plt.plot(self.alphas[optimum_idx], gcv_values[optimum_idx], 'r*')
            alpha = self.alphas[optimum_idx]
            print("alpha: ", alpha)
            plt.title(f"GCV: {alpha}")

        source_mat = self.inverse_operators[optimum_idx].data @ M
        return source_mat[0], optimum_idx

    def regularise_product(self, M, plot=False):
        """ Find optimally regularized inverse solution using the product method [1].
        
        Parameters
        ----------
        M : numpy.ndarray
            The M/EEG data matrix (n_channels, n_timepoints)
        
        Return
        ------
        source_mat : numpy.ndarray
            The inverse solution  (dipoles x time points)
        optimum_idx : int
            The index of the selected (optimal) regularization parameter
        
        References
        ----------
        [1] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
        Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
        problem in EEG source analysis. Journal of neuroengineering and
        rehabilitation, 5(1), 1-33.

        """

        product_values = []

        for inverse_operator in self.inverse_operators:
            x = np.squeeze(inverse_operator.data @ M)

            M_hat = self.leadfield@x
            residual_norm = np.linalg.norm(M_hat - M)
            semi_norm = np.linalg.norm(x)
            product_value = semi_norm * residual_norm 
            product_values.append(product_value)

        optimum_idx = np.argmin(product_values)
        
        if plot:
            plt.figure()
            plt.plot(self.alphas, product_values)
            plt.plot(self.alphas[optimum_idx], product_values[optimum_idx], 'r*')
            alpha = self.alphas[optimum_idx]
            plt.title(f"Product: {alpha}")

        source_mat = self.inverse_operators[optimum_idx].data @ M
        return source_mat[0], optimum_idx

    @staticmethod
    def delete_from_list(a, idc):
        ''' Delete elements of list at idc.'''

        idc = np.sort(idc)[::-1]
        for idx in idc:
            a.pop(idx)
        return a

    def find_corner(self, r_vals, l2_norms):
        ''' Find the corner of the l-curve given by plotting regularization
        levels (r_vals) against norms of the inverse solutions (l2_norms).

        Parameters
        ----------
        r_vals : list
            Levels of regularization
        l2_norms : list
            L2 norms of the inverse solutions per level of regularization.
        
        Return
        ------
        idx : int
            Index at which the L-Curve has its corner.
        '''

        # Normalize l2 norms
        l2_norms /= np.max(l2_norms)

        A = np.array([r_vals[0], l2_norms[0]])
        C = np.array([r_vals[-1], l2_norms[-1]])
        areas = []
        for j in range(1, len(l2_norms)-1):
            B = np.array([r_vals[j], l2_norms[j]])
            AB = self.euclidean_distance(A, B)
            AC = self.euclidean_distance(A, C)
            CB = self.euclidean_distance(C, B)
            area = abs(self.calc_area_tri(AB, AC, CB))
            areas.append(area)
        if len(areas) > 0:
            idx = np.argmax(areas)+1
        else:
            idx = 0
        return idx
    
    @staticmethod
    def select_signal_subspace(data_matrix, rank="auto"):
    
        # Compute the SVD of the data matrix
        U, S, V = np.linalg.svd(data_matrix, full_matrices=False)
        
        if rank == "auto":
            iters = np.arange(len(S))
            n_comp_L = find_corner(deepcopy(iters), deepcopy(S))
            
            # Based on eigenvalue drop-off
            S_ = S/S.max()
            
            n_comp_drop = np.where( abs(np.diff(S_)) < 0.001 )[0]
            if len(n_comp_drop) > 0:
                n_comp_drop = n_comp_drop[0] + 2
            else:
                n_comp_drop = n_comp_L
            
            # Combine the two:
            rank = np.ceil((n_comp_drop + n_comp_L)/2).astype(int)
            
        

        # Select the top `rank` singular values and corresponding singular vectors
        U_subset = U[:, :rank]
        S_subset = S[:rank]
        V_subset = V[:rank, :]

        # Reconstruct a low-rank approximation of the data matrix using the selected singular values and vectors
        data_matrix_approx = U_subset @ np.diag(S_subset) @ V_subset

        return data_matrix_approx

    @staticmethod
    def filter_norms(r_vals, l2_norms):
        ''' Filter l2_norms where they are not monotonically decreasing.

        Parameters
        ----------
        r_vals : [list, numpy.ndarray]
            List or array of r-values
        l2_norms : [list, numpy.ndarray]
            List or array of l2_norms
        
        Return
        ------
        bad_idc : list
            List where indices are increasing

        '''
        diffs = np.diff(l2_norms)
        bad_idc = []
        all_idc = np.arange(len(l2_norms))
        while np.any( diffs>0 ):
            pop_idx = np.where(diffs>0)[0][0]+1
            r_vals = np.delete(r_vals, pop_idx)
            l2_norms = np.delete(l2_norms, pop_idx)
            diffs = np.diff(l2_norms)

            bad_idc.append(all_idc[pop_idx])
            all_idc = np.delete(all_idc, pop_idx)
        return bad_idc

    def prepare_forward(self):
        ''' Prepare leadfield for calculating inverse solutions by applying
        common average referencing and unit norm scaling.

        Parameters
        ----------
        

        Return
        ------
        '''
        # Check whether forward model has free source orientation
        # if yes -> convert to fixed
        if self.forward["source_ori"] == FIFF.FIFFV_MNE_FREE_ORI:
            print("Forward model has free source orientation. This is currently not possible, converting to fixed.")
            # convert to fixed
            self.forward = mne.convert_forward_solution(self.forward, force_fixed=True, verbose=0)
        
        self.leadfield = deepcopy(self.forward["sol"]["data"])
        
        if self.prep_leadfield:
            # self.leadfield -= self.leadfield.mean(axis=0)
            self.leadfield /= np.linalg.norm(self.leadfield, axis=0)
            

    @staticmethod
    def euclidean_distance(A, B):
        ''' Euclidean Distance between two points (A -> B).'''
        return np.sqrt(np.sum((A-B)**2))

    @staticmethod
    def calc_area_tri(AB, AC, CB):
        ''' Calculates area of a triangle given the length of each side.'''
        s = (AB + AC + CB) / 2
        area = (s*(s-AB)*(s-AC)*(s-CB)) ** 0.5
        return area
        
    def source_to_object(self, source_mat):
        ''' Converts the source_mat matrix to the mne.SourceEstimate object.

        Parameters
        ----------
        source_mat : numpy.ndarray
            Source matrix (dipoles, time points)-

        Return
        ------
        stc : mne.SourceEstimate
        
        '''
        # Convert source to mne.SourceEstimate object
        source_model = self.forward['src']
        vertices = [source_model[0]['vertno'], source_model[1]['vertno']]
        tmin = self.tmin
        sfreq = self.obj_info["sfreq"]
        tstep = 1/sfreq
        subject = self.obj_info["subject_info"]

        if type(subject) == dict and "his_id" in subject:
            subject = subject["his_id"]
        # else assume fsaverage as subject id
        else:
            subject = "fsaverage"
        
        stc = mne.SourceEstimate(source_mat, vertices, tmin=tmin, tstep=tstep, subject=subject, verbose=self.verbose)
        return stc
    
    def save(self, path):
        ''' Saves the solver object. 

        Paramters
        ---------
        path : str
            The path to save the solver.
        Return
        ------
        self : BaseSolver
            Function returns itself.
        '''
    
        name = self.name

        # get list of folders in path
        if os.path.exists(path):
            list_of_folders = os.listdir(path)
        else:
            # create path
            os.mkdir(path)
            list_of_folders = []
            
        model_ints = []
        for folder in list_of_folders:
            full_path = os.path.join(path, folder)
            if not os.path.isdir(full_path):
                continue
            if folder.startswith(name):
                new_integer = int(folder.split('_')[-1])
                model_ints.append(new_integer)
        if len(model_ints) == 0:
            model_name = f'\\{name}_0'
        else:
            model_name = f'\\{name}_{max(model_ints)+1}'
        new_path = path+model_name
        os.mkdir(new_path)

        if hasattr(self, "model"):
            # raise AttributeError("tensorflow is no longer supported")
            # Save model only
            self.model.save(new_path)

            # Save rest
            # Delete model since it is not serializable
            self.model = None
            self.generator = None
            if hasattr(self, "history"):
                self.history = None
            
            with open(new_path + '\\instance.pkl', 'wb') as f:
                pkl.dump(self, f, protocol=pkl.HIGHEST_PROTOCOL)
            
            # Attach model again now that everything is saved
            try:
                self.model = tf.keras.models.load_model(new_path, custom_objects={'loss': self.loss})
            except:
                print("Load model did not work using custom_objects. Now trying it without...")
                self.model = tf.keras.models.load_model(new_path)
        else:
            with open(new_path + '\\instance.pkl', 'wb') as f:
                pkl.dump(self, f)
        return self


import numpy as np
import mne
from copy import deepcopy
from .base import BaseSolver, InverseOperator
from scipy.sparse.csgraph import laplacian
from ..util import best_index_residual, thresholding, calc_residual_variance, find_corner

class SolverSMP(BaseSolver):
    ''' Class for the Smooth Matching Pursuit (SMP) inverse solution. Developed
        by Lukas Hecker as a smooth extension of the orthogonal matching pursuit
        algorithm [1,2], 19.10.2022.
    
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Tropp, J. A., & Gilbert, A. C. (2007). Signal recovery from random
    measurements via orthogonal matching pursuit. IEEE Transactions on
    information theory, 53(12), 4655-4666.
    [2] Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing:
    From theory to applications. IEEE Transactions on signal processing, 59(9),
    4053-4085.
     

    '''
    def __init__(self, name="Smooth Matching Pursuit", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.inverse_operators = []
        
        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0).toarray()
        laplace_operator = laplacian(adjacency)
        self.laplace_operator = laplace_operator
        leadfield_smooth = self.leadfield @ abs(laplace_operator)

        leadfield_smooth -= leadfield_smooth.mean(axis=0)
        self.leadfield -= self.leadfield.mean(axis=0)
        self.leadfield_smooth = leadfield_smooth
        self.leadfield_smooth_normed = self.leadfield_smooth / np.linalg.norm(self.leadfield_smooth, axis=0)
        self.leadfield_normed = self.leadfield / np.linalg.norm(self.leadfield, axis=0)
        
        return self

    def apply_inverse_operator(self, mne_obj, K=1, include_singletons=True) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object
        '''
        data = self.unpack_data_obj(mne_obj)

        source_mat = np.stack([self.calc_smp_solution(y, include_singletons=include_singletons) for y in data.T], axis=1)
        stc = self.source_to_object(source_mat)
        return stc
    

    def calc_smp_solution(self, y, include_singletons=True):
        """ Calculates the Orthogonal Matching Pursuit (OMP) inverse solution.
        
        Parameters
        ----------
        y : numpy.ndarray
            The data matrix (channels,).
        include_singletons : bool
            If True -> Include not only smooth patches but also single dipoles.
        

        Return
        ------
        x_hat : numpy.ndarray
            The inverse solution (dipoles,)
        """
        
        n_chans = len(y)
        _, n_dipoles = self.leadfield.shape
        
        y -= y.mean()
        x_hat = np.zeros(n_dipoles)
        omega = np.array([])
        r = deepcopy(y)
        y_hat = self.leadfield@x_hat
        y_hat -= y_hat.mean(axis=0)
        residuals = np.array([np.linalg.norm(y - y_hat), ])
        unexplained_variance = np.array([calc_residual_variance(y_hat, y),])
        # source_norms = np.array([0,])
        x_hats = [deepcopy(x_hat), ]

        for _ in range(n_chans):
            b_smooth = self.leadfield_smooth_normed.T @ r
            b_sparse = self.leadfield_normed.T @ r

            if include_singletons & (abs(b_sparse).max() > abs(b_smooth).max()):  # if sparse is better
                b_sparse_thresh = thresholding(b_sparse, 1)
                new_patch = np.where(b_sparse_thresh != 0)[0]
                
            else: # else if patch is better
                b_smooth_thresh = thresholding(b_smooth, 1)
                new_patch = np.where(self.laplace_operator[b_smooth_thresh!=0][0]!=0)[0]


            omega = np.append(omega, new_patch)
            omega = omega.astype(int)
            x_hat[omega] = np.linalg.pinv(self.leadfield[:, omega]) @ y
            y_hat = self.leadfield@x_hat
            y_hat -= y_hat.mean(axis=0)
            r = y - y_hat

            residuals = np.append(residuals, np.linalg.norm(r))
            unexplained_variance = np.append(unexplained_variance, calc_residual_variance(y_hat, y))
            # source_norms = np.append(source_norms, np.sum(x_hat**2))
            x_hats.append(deepcopy(x_hat))
            if residuals[-1] > residuals[-2]:
                break

        x_hat = best_index_residual(unexplained_variance, x_hats, plot=False)
        
        # x_hat = x_hats[corner_idx]
        return x_hat

class SolverSSMP(BaseSolver):
    ''' Class for the Smooth Simultaneous Matching Pursuit (SSMP) inverse
        solution. Developed by Lukas Hecker as a smooth extension of the
        orthogonal matching pursuit algorithm [1,2], 19.10.2022.
    
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing:
    From theory to applications. IEEE Transactions on signal processing, 59(9),
    4053-4085.

    [2] Donoho, D. L. (2006). Compressed sensing. IEEE Transactions on
    information theory, 52(4), 1289-1306.

    '''
    def __init__(self, name="Smooth Simultaneous Matching Pursuit", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.inverse_operators = []
        
        # Prepare spatial laplacian operator
        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0).toarray()
        laplace_operator = laplacian(adjacency)
        self.laplace_operator = laplace_operator
        
        # Create a leadfield of  smooth sources
        leadfield_smooth = self.leadfield @ abs(laplace_operator)

        leadfield_smooth -= leadfield_smooth.mean(axis=0)
        self.leadfield -= self.leadfield.mean(axis=0)
        self.leadfield_smooth = leadfield_smooth
        self.leadfield_smooth_normed = self.leadfield_smooth / np.linalg.norm(self.leadfield_smooth, axis=0)
        self.leadfield_normed = self.leadfield / np.linalg.norm(self.leadfield, axis=0)
        
        return self

    def apply_inverse_operator(self, mne_obj, include_singletons=True) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        include_singletons : bool
            If True -> Include not only smooth patches but also single dipoles.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object
        '''
        data = self.unpack_data_obj(mne_obj)
        source_mat = self.calc_ssmp_solution(data, include_singletons=include_singletons)
        stc = self.source_to_object(source_mat)
        return stc
    

    def calc_ssmp_solution(self, y, include_singletons=True):
        """ Calculates the Smooth Simultaneous Orthogonal Matching Pursuit (SSMP) inverse solution.
        
        Parameters
        ----------
        y : numpy.ndarray
            The data matrix (channels,).
        include_singletons : bool
            If True -> Include not only smooth patches but also single dipoles.

        Return
        ------
        x_hat : numpy.ndarray
            The inverse solution (dipoles,)
        """

        n_chans, n_time = y.shape
        max_iter = int(n_chans/2)
        _, n_dipoles = self.leadfield.shape
        
        y -= y.mean(axis=0)

        x_hat = np.zeros((n_dipoles, n_time))
        x_hats = [deepcopy(x_hat)]
        residuals = np.array([np.linalg.norm(y - self.leadfield@x_hat), ])
        source_norms = np.array([0,])

        R = deepcopy(y)
        omega = np.array([])

        y_hat = self.leadfield @ x_hat
        y_hat -= y_hat.mean(axis=0)
        residuals = np.array([np.linalg.norm(y - y_hat), ])
        x_hats = [deepcopy(x_hat), ]
        q = 1
        
        for _ in range(max_iter):
            b_n_smooth = np.linalg.norm(self.leadfield_smooth_normed.T @ R, axis=1, ord=q)
            b_n_sparse = np.linalg.norm(self.leadfield_normed.T @ R, axis=1, ord=q)

            if include_singletons & (abs(b_n_sparse).max() > abs(b_n_smooth).max()):  # if sparse is better
                b_n_sparse_thresh = thresholding(b_n_sparse, 1)
                new_patch = np.where(b_n_sparse_thresh != 0)[0]
                
            else: # else if patch is better
                b_n_smooth_thresh = thresholding(b_n_smooth, 1)
                new_patch = np.where(self.laplace_operator[b_n_smooth_thresh!=0][0]!=0)[0]


            omega = np.append(omega, new_patch)
            omega = omega.astype(int)
            x_hat[omega] = np.linalg.pinv(self.leadfield[:, omega]) @ y
            
            y_hat = self.leadfield@x_hat
            y_hat -= y_hat.mean(axis=0)
            R = y - y_hat

            residuals = np.append(residuals, np.linalg.norm(y - y_hat))
            source_norms = np.append(source_norms, np.sum(x_hat**2))
            x_hats.append(deepcopy(x_hat))

            if residuals[-1] > residuals[-2]:
                break



        # Model selection (Regularisation)
        x_hat = best_index_residual(residuals, x_hats, plot=False)

        return x_hat
    
class SolverSubSMP(BaseSolver):
    ''' Class for the Subspace Smooth Matching Pursuit (SubSMP) inverse solution. Developed
        by Lukas Hecker as a smooth extension of the orthogonal matching pursuit
        algorithm [1], 19.10.2022.
    
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing:
    From theory to applications. IEEE Transactions on signal processing, 59(9),
    4053-4085.

    '''
    def __init__(self, name="Subspace Smooth Matching Pursuit", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.inverse_operators = []
        
        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0).toarray()
        
        # Create spatial laplacian operator
        laplace_operator = laplacian(adjacency)
        self.laplace_operator = laplace_operator
        
        # Create leadfield of smooth sources
        leadfield_smooth = self.leadfield @ abs(laplace_operator)
        leadfield_smooth -= leadfield_smooth.mean(axis=0)
        self.leadfield -= self.leadfield.mean(axis=0)
        self.leadfield_smooth = leadfield_smooth
        self.leadfield_smooth_normed = self.leadfield_smooth / np.linalg.norm(self.leadfield_smooth, axis=0)
        self.leadfield_normed = self.leadfield / np.linalg.norm(self.leadfield, axis=0)
        
        return self

    def apply_inverse_operator(self, mne_obj,include_singletons=True) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object
        '''
        data = self.unpack_data_obj(mne_obj)
        source_mat = self.calc_subsmp_solution(data, include_singletons=include_singletons)
        stc = self.source_to_object(source_mat)
        return stc

    def calc_subsmp_solution(self, y, include_singletons=True, var_thresh=1):
        ''' Calculate the Subspace Smooth Matching Pursuit (SubSMP) solution.

        Parameters
        ----------
        y : numpy.ndarray
            The M/EEG data Matrix
        include_singletons : bool
            If True -> include single dipoles as candidates, 
            else include only smooth patches
        var_thresh : float
            Threshold how much variance will be explained in the data

        '''
        _, n_time = y.shape
        n_dipoles = self.leadfield.shape[1]
        
        y -= y.mean(axis=0)
        u, s, _ = np.linalg.svd(y)
        
        # Select components
        idx = find_corner(np.arange(len(s)), s) + 1
        print(f"Selected {idx} components")
        # Kaiser-Guttmann:
        # thr = 1  # 1=kaiser-guttmann, else: np.e**(-16)
        # idx = np.where((s*len(s) / s.sum()) > thr)[0][-1] + 1

        # Find matching sources to each of the selected eigenvectors (components)
        topos = u[:, :idx]
        x_hats = []
        for topo in topos.T:
            x_hats.append( self.calc_smp_solution(topo, include_singletons=include_singletons, var_thresh=var_thresh) )
        
        # Extract all active dipoles and patches:
        omegas = np.unique(np.where((np.stack(x_hats, axis=0)**2).sum(axis=0) != 0)[0])
        
        # Calculate minimum-norm solution at active dipoles and patches
        x_hat = np.zeros((n_dipoles, n_time))

        ## Use different y without unnecessary components here:
        x_hat[omegas, :] = np.linalg.pinv(self.leadfield[:, omegas]) @ y
        
        # import matplotlib.pyplot as plt
        # plt.figure()
        # plt.plot(np.arange(len(s)), s, '*k')
        # plt.plot(np.arange(len(s))[idx], s[idx], 'or')
        

        return x_hat

    def calc_smp_solution(self, y, include_singletons=True, var_thresh=1):
        """ Calculates the Orthogonal Matching Pursuit (OMP) inverse solution.
        
        Parameters
        ----------
        y : numpy.ndarray
            The data matrix (channels,).
        include_singletons : bool
            If True -> Include not only smooth patches but also single dipoles.
        var_thresh : float
            Threshold how much variance will be explained in the data
            
        Return
        ------
        x_hat : numpy.ndarray
            The inverse solution (dipoles,)
        """
        n_chans = len(y)
        _, n_dipoles = self.leadfield.shape
        
        y -= y.mean()
        x_hat = np.zeros(n_dipoles)
        omega = np.array([])
        r = deepcopy(y)
        y_hat = self.leadfield@x_hat
        y_hat -= y_hat.mean(axis=0)
        residuals = np.array([np.linalg.norm(y - y_hat), ])
        unexplained_variance = np.array([calc_residual_variance(y_hat, y),])
        # source_norms = np.array([0,])
        x_hats = [deepcopy(x_hat), ]

        for _ in range(n_chans):
            b_smooth = self.leadfield_smooth_normed.T @ r
            b_sparse = self.leadfield_normed.T @ r

            if include_singletons & (abs(b_sparse).max() > abs(b_smooth).max()):  # if sparse is better
                b_sparse_thresh = thresholding(b_sparse, 1)
                new_patch = np.where(b_sparse_thresh != 0)[0]
                
            else: # else if patch is better
                b_smooth_thresh = thresholding(b_smooth, 1)
                new_patch = np.where(self.laplace_operator[b_smooth_thresh!=0][0]!=0)[0]


            omega = np.append(omega, new_patch)
            omega = omega.astype(int)
            x_hat[omega] = np.linalg.pinv(self.leadfield[:, omega]) @ y
            y_hat = self.leadfield@x_hat
            y_hat -= y_hat.mean(axis=0)
            r = y - y_hat

            residuals = np.append(residuals, np.linalg.norm(r))
            unexplained_variance = np.append(unexplained_variance, calc_residual_variance(y_hat, y))
            # source_norms = np.append(source_norms, np.sum(x_hat**2))
            x_hats.append(deepcopy(x_hat))
            if residuals[-1] > residuals[-2]:
                break

        if unexplained_variance[1] > var_thresh:
            x_hat = best_index_residual(unexplained_variance, x_hats, plot=False)
        else:
            x_hat = x_hats[1]
        
        # x_hat = x_hats[corner_idx]
        return x_hat
    
class SolverISubSMP(BaseSolver):
    ''' Class for the Subspace Smooth Matching Pursuit (SubSMP) inverse solution. Developed
        by Lukas Hecker as a smooth extension of the orthogonal matching pursuit
        algorithm [1], 19.10.2022.
    
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    [1] Duarte, M. F., & Eldar, Y. C. (2011). Structured compressed sensing:
    From theory to applications. IEEE Transactions on signal processing, 59(9),
    4053-4085.

    '''
    def __init__(self, name="Iterative Subspace Smooth Matching Pursuit", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.inverse_operators = []
        
        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0).toarray()
        laplace_operator = laplacian(adjacency)
        self.laplace_operator = laplace_operator
        leadfield_smooth = self.leadfield @ abs(laplace_operator)

        leadfield_smooth -= leadfield_smooth.mean(axis=0)
        self.leadfield -= self.leadfield.mean(axis=0)
        self.leadfield_smooth = leadfield_smooth
        self.leadfield_smooth_normed = self.leadfield_smooth / np.linalg.norm(self.leadfield_smooth, axis=0)
        self.leadfield_normed = self.leadfield / np.linalg.norm(self.leadfield, axis=0)
        
        return self

    def apply_inverse_operator(self, mne_obj, include_singletons=True) -> mne.SourceEstimate:
        ''' Apply the inverse operator.
        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.

        Return
        ------
        stc : mne.SourceEstimate
            The mne Source Estimate object
        '''
        data = self.unpack_data_obj(mne_obj)
        source_mat = self.calc_subsmp_solution(data, include_singletons=include_singletons)
        stc = self.source_to_object(source_mat)
        return stc

    def calc_subsmp_solution(self, y, include_singletons=True, var_thresh=0.1):
        ''' Calculate the Subspace Smooth Matching Pursuit (SubSMP) solution.
        '''
        _, n_time = y.shape
        n_dipoles = self.leadfield.shape[1]
        
        y -= y.mean(axis=0)


        # Find matching sources to each of the selected eigenvectors (components)
 
        x_hat = self.calc_isubsmp_solution(y, include_singletons=include_singletons, var_thresh=var_thresh)

        return x_hat

    def calc_isubsmp_solution(self, y, include_singletons=True, var_thresh=1):
        """ Calculates the Orthogonal Matching Pursuit (OMP) inverse solution.
        
        Parameters
        ----------
        y : numpy.ndarray
            The data matrix (channels, n_time).
        include_singletons : bool
            If True -> Include not only smooth patches but also single dipoles.

        Return
        ------
        x_hat : numpy.ndarray
            The inverse solution (dipoles,)
        """
        n_chans, n_time = y.shape
        _, n_dipoles = self.leadfield.shape
        
        y -= y.mean(axis=0)

        # print(y.shape)
        x_hat = np.zeros((n_dipoles, n_time))
        omega = np.array([])
        r = deepcopy(y)
        r_component = np.linalg.svd(r, full_matrices=False)[0][:, 0]
        y_hat = self.leadfield@x_hat
        y_hat -= y_hat.mean(axis=0)
        residuals = np.array([np.linalg.norm(y - y_hat), ])
        unexplained_variance = np.array([calc_residual_variance(y_hat, y),])
        # source_norms = np.array([0,])
        x_hats = [deepcopy(x_hat), ]

        for _ in range(n_chans):
            b_smooth = self.leadfield_smooth_normed.T @ r_component
            b_sparse = self.leadfield_normed.T @ r_component

            if include_singletons & (abs(b_sparse).max() > abs(b_smooth).max()):  # if sparse is better
                b_sparse_thresh = thresholding(b_sparse, 1)
                new_patch = np.where(b_sparse_thresh != 0)[0]
                
            else: # else if patch is better
                b_smooth_thresh = thresholding(b_smooth, 1)
                new_patch = np.where(self.laplace_operator[b_smooth_thresh!=0][0]!=0)[0]


            omega = np.append(omega, new_patch)
            omega = omega.astype(int)
            x_hat[omega] = np.linalg.pinv(self.leadfield[:, omega]) @ y
            y_hat = self.leadfield@x_hat
            y_hat -= y_hat.mean(axis=0)
            r = y - y_hat
            r_component = np.linalg.svd(r, full_matrices=False)[0][:, 0]

            residuals = np.append(residuals, np.linalg.norm(r))
            unexplained_variance = np.append(unexplained_variance, calc_residual_variance(y_hat, y))
            # source_norms = np.append(source_norms, np.sum(x_hat**2))
            x_hats.append(deepcopy(x_hat))
            if residuals[-1] > residuals[-2] or unexplained_variance[-1]<var_thresh:
                break



        
        # Remove where residual is below threshold:
        # keep_idc = np.where(unexplained_variance>=residual_thresh*100)[0]
        # if len(keep_idc) == 1:
        #     keep_idc = np.append(keep_idc, 1)
        
        # print(keep_idc)
        # unexplained_variance = unexplained_variance[keep_idc]
        # x_hats = [x_hats[idx] for idx in keep_idc]
        # if unexplained_variance[-1] > var_thresh:
        #     x_hat = best_index_residual(unexplained_variance, x_hats, plot=False)
        # else:
        #     x_hat = x_hats[-1]
        
        # x_hat = x_hats[corner_idx]
        return x_hat


import numpy as np
import mne
from copy import deepcopy
from scipy.sparse.csgraph import laplacian
from scipy.sparse import coo_matrix
from .base import BaseSolver, InverseOperator

class SolverBCS(BaseSolver):
    ''' Class for the Bayesian Compressed Sensing (BCS) inverse solution [1].
    
    Attributes
    ----------
    
    
    References
    ----------
    [1] Ji, S., Xue, Y., & Carin, L. (2008). Bayesian compressive sensing. IEEE
    Transactions on signal processing, 56(6), 2346-2356.

    '''
    def __init__(self, name="Bayesian Compressed Sensing", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha="auto", **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        self.leadfield_norm = self.leadfield

        return self

    def apply_inverse_operator(self, mne_obj, max_iter=100, alpha_0=0.01, eps=1e-16) -> mne.SourceEstimate:
        ''' Apply the inverse operator.

        Parameters
        ----------
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        max_iter : int
            Maximum number of iterations
        alpha_0 : float
            Regularization parameter
        eps : float
            Epsilon, used to avoid division by zero.
        
        Return
        ------
        stc : mne.SourceEstimate
            The SourceEstimate data structure containing the inverse solution.

        '''
        data = self.unpack_data_obj(mne_obj)
        source_mat = self.calc_bcs_solution(data, max_iter=max_iter, alpha_0=alpha_0, eps=eps)
        stc = self.source_to_object(source_mat)
        return stc
    
    def calc_bcs_solution(self, y, max_iter=100, alpha_0=0.01, eps=1e-16):
        ''' This function computes the BCS inverse solution.

        Parameters
        ----------
        y : numpy.ndarray
            The M/EEG data matrix (n_channels, n_timepoints)
        max_iter : int
            Maximum number of iterations
        alpha_0 : float
            Regularization parameter
        eps : float
            Epsilon, used to avoid division by zero.
        
        Return
        ------
        x_hat : numpy.ndarray
            The source estimate.
        '''

        alpha_0 = np.clip(alpha_0, a_min=1e-6, a_max=None)
        n_chans, _ = y.shape
        n_dipoles = self.leadfield_norm.shape[1]
        
        # preprocessing
        y -= y.mean(axis=0)
        
        alphas = np.ones(n_dipoles)
        D = np.diag(alphas)

        LLT = self.leadfield_norm.T @ self.leadfield_norm
        sigma = np.linalg.inv(alpha_0 * LLT + D)
        mu = alpha_0 * sigma @ self.leadfield_norm.T @ y
        proj_norm = self.leadfield_norm.T @ y
        proj = self.leadfield.T @ y
        
        residual_norms = [1e99]
        x_hats = []
        for i in range(max_iter):
            gammas = np.array([1 - alphas[ii] * sigma[ii,ii] for ii in range(n_dipoles)])
            gammas[np.isnan(gammas)] = 0

            
            alphas = gammas / np.linalg.norm(mu**2, axis=1)
            alpha_0 = 1 / ( np.linalg.norm(y - self.leadfield_norm @ mu) / (n_chans - gammas.sum()) )
            D = np.diag(alphas) + eps
            sigma = np.linalg.inv(alpha_0 * LLT + D)
            mu = alpha_0 * sigma @ proj_norm

            # var = 1/alpha_0
            # try:
            #     D_inv = np.linalg.inv(D)
            # except:
            #     break
            # C = var**2 * I + self.leadfield @ D_inv @ self.leadfield.T
            
            Gamma = np.diag(gammas)
            x_hat = Gamma @ proj
            residual_norm = np.linalg.norm(y - self.leadfield @ x_hat)
            # print(residual_norm)
            if residual_norm > residual_norms[-1]:
                x_hat = x_hats[-1]
                # print(f"Stopping after {i} iterations.")
                break
            residual_norms.append(residual_norm)
            x_hats.append(x_hat)
            # print("residual_norm: ", residual_norm)

        return x_hat

class SolverGammaMAP(BaseSolver):
    ''' Class for the Gamma Maximum A Posteriori (Gamma-MAP) inverse solution [1].
    
    Attributes
    ----------
    
    References
    ----------
    Wipf, D., & Nagarajan, S. (2009). A unified Bayesian framework for MEG/EEG
    source imaging. NeuroImage, 44(3), 947-966.

    '''
    def __init__(self, name="Gamma-MAP", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha="auto", smoothness_prior=False,
                              max_iter=100, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : str/ float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        smoothness_prior : bool

        
        
        Return
        ------
        self : object returns itself for convenience

        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        leadfield = self.leadfield
        n_chans, n_dipoles = leadfield.shape
        data = self.unpack_data_obj(mne_obj)

        if smoothness_prior:
            adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0)
            self.gradient = laplacian(adjacency).toarray().astype(np.float32)
            self.sigma_s = np.identity(n_dipoles) @ abs(self.gradient)
        else:
            self.gradient = None
            self.sigma_s = np.identity(n_dipoles)
        
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = self.make_gamma_map_inverse_operator(data, alpha, max_iter=max_iter)
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

    
    def make_gamma_map_inverse_operator(self, B, alpha, max_iter=100):
        ''' Computes the gamma MAP inverse operator based on the M/EEG data.
        
        Parameters
        ----------
        B : numpy.ndarray
            The M/EEG data matrix (channels, time points).
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        
        Return
        ------
        inverse_operator : numpy.ndarray
            The inverse operator which can be used to compute inverse solutions from new data.

        '''
        L = deepcopy(self.leadfield)
        db, n = B.shape
        ds = L.shape[1]

        # Ensure Common average reference
        B -= B.mean(axis=0)
        L -= L.mean(axis=0)
        
        # Data Covariance Matrix
        gammas = np.ones(ds)
        sigma_e = alpha * np.identity(db)  
        
        sigma_b = sigma_e + L @ self.sigma_s @ L.T
        sigma_b_inv = np.linalg.inv(sigma_b)
        
        for k in range(max_iter):
            old_gammas = deepcopy(gammas)
            
            # according to equation (30)
            term_1 = (gammas/np.sqrt(n)) * np.sqrt(np.sum((L.T @ sigma_b_inv @ B )**2, axis=1))
            term_2 = 1 / np.sqrt(np.diagonal((L.T @ sigma_b_inv @ L )))

            gammas = term_1 * term_2
            if np.linalg.norm(gammas) == 0:
                gammas = old_gammas
                break

        gammas_final = gammas / gammas.max()
        sigma_s_hat = np.diag(gammas_final) @ self.sigma_s  #  np.array([gammas_final[i] * C[i] for i in range(ds)])
        inverse_operator = sigma_s_hat @ L.T @ np.linalg.inv(sigma_e + L @ sigma_s_hat @ L.T)
        
        # This way the inverse operator would be applied to M/EEG matrix B:
        # S = inverse_operator @ B

        return inverse_operator
        
    @staticmethod
    def frob(x):
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        return np.sqrt(np.trace(x@x.T))

class SolverSourceMAP(BaseSolver):
    ''' Class for the Source Maximum A Posteriori (Source-MAP) inverse solution [1].
    
    Attributes
    ----------
    
    References
    ----------
    Wipf, D., & Nagarajan, S. (2009). A unified Bayesian framework for MEG/EEG
    source imaging. NeuroImage, 44(3), 947-966.

    '''
    def __init__(self, name="Source-MAP", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha="auto", smoothness_prior=False,
                              max_iter=100, p=0.5, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        p : 0 < p < 2 
            Hyperparameter which controls sparsity. Default: p = 0.5
        
        Return
        ------
        self : object returns itself for convenience

        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)
        leadfield = self.leadfield
        n_chans, n_dipoles = leadfield.shape

        if smoothness_prior:
            adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0)
            self.gradient = laplacian(adjacency).toarray().astype(np.float32)
            self.sigma_s = np.identity(n_dipoles) @ abs(self.gradient)
        else:
            self.gradient = None
            self.sigma_s = np.identity(n_dipoles)

        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = self.make_source_map_inverse_operator(data, alpha, max_iter=max_iter, p=p)
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

    
    def make_source_map_inverse_operator(self, B, alpha, max_iter=100, p=0.5):
        ''' Computes the source MAP inverse operator based on the M/EEG data.
        
        Parameters
        ----------
        B : numpy.ndarray
            The M/EEG data matrix (channels, time points).
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        p : 0 < p < 2 
            Hyperparameter which controls sparsity. Default: p = 0.5

        
        Return
        ------
        inverse_operator : numpy.ndarray
            The inverse operator which can be used to compute inverse solutions from new data.

        '''

        L = deepcopy(self.leadfield)
        db, n = B.shape
        ds = L.shape[1]

        # Ensure Common average reference
        B -= B.mean(axis=0)
        L -= L.mean(axis=0)
        L /= np.linalg.norm(L, axis=0)
 
        # Data Covariance Matrix
        # Cb = B @ B.T
        gammas = np.ones(ds)
        sigma_e = alpha * np.identity(db)

        sigma_b = sigma_e + L @ self.sigma_s @ L.T
        sigma_b_inv = np.linalg.inv(sigma_b)
        
        for k in range(max_iter):
            # print(k)
            old_gammas = deepcopy(gammas)
            
            gammas = ((1/n) * np.sqrt(np.sum(( np.diag(gammas) @ L.T @ sigma_b_inv @ B )**2, axis=1)))**((2-p)/2)

            if np.linalg.norm(gammas) == 0:
                gammas = old_gammas
                break
            # gammas /= np.linalg.norm(gammas)

        gammas_final = gammas / gammas.max()
        sigma_s_hat = np.diag(gammas_final) @ self.sigma_s  #  np.array([gammas_final[i] * C[i] for i in range(ds)])
        inverse_operator = sigma_s_hat @ L.T @ np.linalg.inv(sigma_e + L @ sigma_s_hat @ L.T)

        # This way the inverse operator would be applied to M/EEG matrix B:
        # S = inverse_operator @ B

        return inverse_operator
        
    @staticmethod
    def frob(x):
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        return np.sqrt(np.trace(x@x.T))

class SolverGammaMAPMSP(BaseSolver):
    ''' Class for the Gamma Maximum A Posteriori (Gamma-MAP) inverse solution
    using multiple sparse priors (MSP).
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    
    References
    ----------
    Wipf, D., & Nagarajan, S. (2009). A unified Bayesian framework for MEG/EEG
    source imaging. NeuroImage, 44(3), 947-966.
    
    '''
    def __init__(self, name="Gamma-MAP-MSP", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha="auto", 
                              max_iter=100, p=0.5, smoothness_order=1, verbose=0, 
                              **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA-like solution.
        p : 0 < p < 2 
            Hyperparameter which controls sparsity. Default: p = 0
        smoothness_order : int
            Controls the smoothness prior. The higher this integer, the higher
            the pursued smoothness of the inverse solution.

        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)
        leadfield = self.leadfield
        n_chans, _ = leadfield.shape
        
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = self.make_source_map_inverse_operator(data, alpha, 
                                                                    max_iter=max_iter, p=p, 
                                                                    smoothness_order=smoothness_order)
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

    def make_source_map_inverse_operator(self, B, alpha, max_iter=100, p=0.5, smoothness_order=1):
        ''' Computes the source MAP inverse operator based on the M/EEG data.
        
        Parameters
        ----------
        B : numpy.ndarray
            The M/EEG data matrix (channels, time points).
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        p : 0 < p < 2 
            Hyperparameter which controls sparsity. Default: p = 0.5
        smoothness_order : int
            Controls the smoothness prior. The higher this integer, the higher
            the pursued smoothness of the inverse solution.
        
        Return
        ------
        inverse_operator : numpy.ndarray
            The inverse operator which can be used to compute inverse solutions from new data.

        '''
        L = deepcopy(self.leadfield)
        db, n = B.shape
        ds = L.shape[1]

        # Ensure Common average reference
        B -= B.mean(axis=0)
        L -= L.mean(axis=0)
        
 
        # Data Covariance Matrix
        # Cb = B @ B.T
        L_smooth, gradient = self.get_smooth_prior_cov(L, smoothness_order)
        gammas = np.ones(ds)
        sigma_e = alpha * np.identity(db)  
        sigma_s = np.identity(ds) # identity leads to weighted minimum L2 Norm-type solution
        sigma_b = sigma_e + L_smooth @ sigma_s @ L_smooth.T
        sigma_b_inv = np.linalg.inv(sigma_b)
        
        for k in range(max_iter):
            # print(k)
            old_gammas = deepcopy(gammas)
            
            # gammas = ((1/n) * np.sqrt(np.sum(( np.diag(gammas) @ L_smooth.T @ sigma_b_inv @ B )**2, axis=1)))**((2-p)/2)
            
            term_1 = (gammas/np.sqrt(n)) * np.sqrt(np.sum((L_smooth.T @ sigma_b_inv @ B )**2, axis=1))
            term_2 = 1 / np.sqrt(np.diagonal((L_smooth.T @ sigma_b_inv @ L_smooth )))
            gammas = term_1 * term_2

            if np.linalg.norm(gammas) == 0:
                gammas = old_gammas
                break
            # print(gammas.min(), gammas.max())
            # gammas /= np.linalg.norm(gammas)
        
        # Smooth gammas according to smooth priors
        gammas_final = abs(gammas@gradient)
        gammas_final = gammas / gammas.max()

        sigma_s_hat = np.diag(gammas_final) @ sigma_s
        inverse_operator = sigma_s_hat @ L.T @ np.linalg.inv(sigma_e + L @ sigma_s_hat @ L.T)
        # S = inverse_operator @ B
        return inverse_operator
        
    @staticmethod
    def frob(x):
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        return np.sqrt(np.trace(x@x.T))
    
    def get_smooth_prior_cov(self, L, smoothness_order):
        ''' Create a smooth prior on the covariance matrix.
        
        Parameters
        ----------
        L : numpy.ndarray
            Leadfield matrix (channels, dipoles)
        smoothness_order : int
            The higher the order, the smoother the prior.

        Return
        ------
        L : numpy.ndarray
            The smoothed Leadfield matrix (channels, dipoles)
        gradient : numpy.ndarray
            The smoothness gradient (laplacian matrix)

        '''
        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0)
        gradient = laplacian(adjacency).toarray().astype(np.float32)
        
        for _ in range(smoothness_order):
            gradient = gradient @ gradient
        L = L @ abs(gradient)
        L -= L.mean(axis=0)
        # L /= np.linalg.norm(L, axis=0)
        return L, gradient

class SolverSourceMAPMSP(BaseSolver):
    ''' Class for the Source Maximum A Posteriori (Source-MAP) inverse solution
    using multiple sparse priors [1]. The method is conceptually similar to [2],
    but formally not equal.
    
    Attributes
    ----------

    
    References
    ----------
    [1] Wipf, D., & Nagarajan, S. (2009). A unified Bayesian framework for
    MEG/EEG source imaging. NeuroImage, 44(3), 947-966. 
    
    [2] Friston, K., Harrison, L., Daunizeau, J., Kiebel, S., Phillips, C.,
    Trujillo-Barreto, N., ... & Mattout, J. (2008). Multiple sparse priors for
    the M/EEG inverse problem. NeuroImage, 39(3), 1104-1120.
    
    '''
    def __init__(self, name="Source-MAP-MSP", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, mne_obj, *args, alpha="auto", 
                              max_iter=100, p=0.5, smoothness_order=1, verbose=0, 
                              **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        mne_obj : [mne.Evoked, mne.Epochs, mne.io.Raw]
            The MNE data object.
        alpha : float
            The regularization parameter.
        p : 0 < p < 2 
            Hyperparameter which controls sparsity. Default: p = 0.5
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        smoothness_order : int
            Controls the smoothness prior. The higher this integer, the higher
            the pursued smoothness of the inverse solution.

        Return
        ------
        self : object returns itself for convenience

        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        data = self.unpack_data_obj(mne_obj)
        leadfield = self.leadfield
        n_chans, _ = leadfield.shape
        
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = self.make_source_map_inverse_operator(data, alpha, 
                                                                    max_iter=max_iter, p=p, 
                                                                    smoothness_order=smoothness_order)
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self
    
    def make_source_map_inverse_operator(self, B, alpha, max_iter=100, p=0.5, smoothness_order=1):
        ''' Computes the source MAP inverse operator based on the M/EEG data.
        
        Parameters
        ----------
        B : numpy.ndarray
            The M/EEG data matrix (channels, time points).
        alpha : float
            The regularization parameter.
        max_iter : int
            Maximum numbers of iterations to find the optimal hyperparameters.
            max_iter = 1 corresponds to sLORETA.
        p : 0 < p < 2 
            Hyperparameter which controls sparsity. Default: p = 0.5
        smoothness_order : int
            Controls the smoothness prior. The higher this integer, the higher
            the pursued smoothness of the inverse solution.
        
        Return
        ------
        inverse_operator : numpy.ndarray
            The inverse operator which can be used to compute inverse solutions from new data.

        '''

        L = deepcopy(self.leadfield)
        db, n = B.shape
        ds = L.shape[1]

        # Ensure Common average reference
        B -= B.mean(axis=0)
        L -= L.mean(axis=0)
        
 
        # Data Covariance Matrix
        # Cb = B @ B.T
        L_smooth, gradient = self.get_smooth_prior_cov(L, smoothness_order)
        gammas = np.ones(ds)
        sigma_e = alpha * np.identity(db)  
        sigma_s = np.identity(ds) # identity leads to weighted minimum L2 Norm-type solution
        sigma_b = sigma_e + L_smooth @ sigma_s @ L_smooth.T
        sigma_b_inv = np.linalg.inv(sigma_b)
        
        for k in range(max_iter):
            # print(k)
            old_gammas = deepcopy(gammas)
            
            gammas = ((1/n) * np.sqrt(np.sum(( np.diag(gammas) @ L_smooth.T @ sigma_b_inv @ B )**2, axis=1)))**((2-p)/2)

            if np.linalg.norm(gammas) == 0:
                gammas = old_gammas
                break
            # print(gammas.min(), gammas.max())
            # gammas /= np.linalg.norm(gammas)
        
        # Smooth gammas according to smooth priors
        gammas_final = abs(gammas@gradient)
        gammas_final = gammas / gammas.max()

        sigma_s_hat = np.diag(gammas_final) @ sigma_s  #  np.array([gammas_final[i] * C[i] for i in range(ds)])
        inverse_operator = sigma_s_hat @ L.T @ np.linalg.inv(sigma_e + L @ sigma_s_hat @ L.T)
        # S = inverse_operator @ B
        return inverse_operator
        
    @staticmethod
    def frob(x):
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        return np.sqrt(np.trace(x@x.T))
    
    def get_smooth_prior_cov(self, L, smoothness_order):
        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0)
        gradient = laplacian(adjacency).toarray().astype(np.float32)
        
        for i in range(smoothness_order):
            gradient = gradient @ gradient
        L = L @ abs(gradient)
        L -= L.mean(axis=0)
        # L /= np.linalg.norm(L, axis=0)
        return L, gradient


import numpy as np
from copy import deepcopy
import mne

from invert.util.util import pos_from_forward

def focuss(stc, evoked, forward, alpha=0.01, max_iter=10, verbose=0):
    ''' FOCUSS algorithm.

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source Estimate object.
    evoked : mne.EvokedArray
        Evoked EEG data object
    forward : mne.Forward
        The forward model
    alpha : float,
        Regularization parameter
    verbose : int
        Controls verbosity of the program
    
    Return
    ------
    stc_focuss : mne.SourceEstimate
        The new focussed source estimate
    '''

    leadfield = forward['sol']['data']
    n_chans, _ = leadfield.shape
    D = stc.data
    M = evoked.data
    w_i = np.diag(1/np.linalg.norm(leadfield, axis=0))
        
    D_FOCUSS = np.zeros(D.shape)
    if verbose:
        print("FOCUSS:\n")
    for t in range(D.shape[1]):
        if verbose > 0:
            print(f"Time step {t}/{D.shape[1]}")
        D_Last = deepcopy(D)
            
        W_i = np.diag(D[:, t])
        
        for i in range(max_iter):
            if verbose > 0:
                print(f'Iteration {i}')
            WWL = W_i @ W_i.T @ leadfield.T
            D_FOCUSS_t = WWL @ np.linalg.inv(leadfield @ WWL + alpha * np.identity(n_chans)) @ M[:, t][:, np.newaxis]
            W_i = w_i @ W_i @ np.diag(D_FOCUSS_t[:, 0])

            if np.linalg.norm(D_FOCUSS_t) == 0:
                D_FOCUSS_t = D_Last
                if verbose:
                    print(f"converged at repetition {i+1}")
                D_FOCUSS[:, t] = D_FOCUSS_t[:, 0]
                break
            else:
                D_Last = deepcopy(D_FOCUSS_t)

    stc_focuss = stc.copy()
    stc_focuss.data = D_FOCUSS
    return stc_focuss

def s_focuss(stc, evoked, forward, alpha=0.01, percentile=0.01, max_iter=10, verbose=0):
    ''' Shrinking FOCUSS algorithm. Based on Grech et al. (2008)

    Parameters
    ----------
    stc : mne.SourceEstimate
        Source Estimate object.
    evoked : mne.EvokedArray
        Evoked EEG data object
    forward : mne.Forward
        The forward model
    alpha : float,
        Regularization parameter
    verbose : int
        Controls verbosity of the program
    
    Return
    ------
    stc_focuss : mne.SourceEstimate
        The new focussed source estimate
    
    References
    ----------
    [1] Grech, R., Cassar, T., Muscat, J., Camilleri, K. P., Fabri, S. G.,
    Zervakis, M., ... & Vanrumste, B. (2008). Review on solving the inverse
    problem in EEG source analysis. Journal of neuroengineering and
    rehabilitation, 5(1), 1-33.

    '''

    leadfield = forward['sol']['data']
    adjacency = mne.spatial_src_adjacency(forward['src'], verbose=verbose).toarray()
    pos = pos_from_forward(forward)
    n_chans, n_dipoles = leadfield.shape
    D = stc.data
    M = evoked.data
    
    w_i = np.diag(1/np.linalg.norm(leadfield, axis=0))
        
    D_FOCUSS = np.zeros(D.shape)
    if verbose:
        print("Shrinking FOCUSS:\n")
    for t in range(D.shape[1]):
        if verbose > 0:
            print(f"Time step {t}/{D.shape[1]}")
        D_Last = deepcopy(D)
        do_smoothing = True
        
        # Initial weighting matrix
        W_i = np.diag(D[:, t])
        WWL = W_i @ W_i.T @ leadfield.T
        sparsities = [1e9,]
        for iter in range(max_iter):
            # if verbose > 0:
                # print(f'\tIteration {iter+1}')
            
            D_FOCUSS_t = WWL @ np.linalg.inv(leadfield @ WWL + alpha * np.identity(n_chans)) @ M[:, t][:, np.newaxis]
            # Smoothing operation (step 4 and 5)
            if do_smoothing:
                D_FOCUSS_t_smooth, sparsity = smooth(D_FOCUSS_t, adjacency, percentile=percentile)
                # leadfield[:, D_FOCUSS_t_smooth[:, 0]==0] *= 0
            else:
                D_FOCUSS_t_smooth = D_FOCUSS_t
            
            
            if do_smoothing:
                if sparsity<n_chans or sparsity>sparsities[-1]:
                    do_smoothing = False
                    # print("\tShrinking is done")
                else:
                    # print(f"sparsity at {sparsity}")
                    sparsities.append(sparsity)
                    D_Last = deepcopy(D_FOCUSS_t_smooth)
                    W_i = w_i @ W_i @ np.diag(D_FOCUSS_t_smooth[:, 0])
            else:
                # Continue with normal FOCUSS
                if np.linalg.norm(D_FOCUSS_t) == 0:
                    D_FOCUSS_t = D_Last
                    # if verbose:
                        # print(f"\tconverged at repetition {iter+1}")
                    D_FOCUSS[:, t] = D_FOCUSS_t[:, 0]
                    break
                else:
                    D_Last = deepcopy(D_FOCUSS_t)
                    W_i = w_i @ W_i @ np.diag(D_FOCUSS_t_smooth[:, 0])
                
                    

    stc_focuss = stc.copy()
    stc_focuss.data = D_FOCUSS
    return stc_focuss

def smooth(D_FOCUSS_t, adjacency, percentile=0.01):

    D_FOCUSS_t_smoothed = deepcopy(D_FOCUSS_t)
    n_dipoles = adjacency.shape[0]

    maximum_value = abs(D_FOCUSS_t).max()
    prominent_idc = np.where(abs(D_FOCUSS_t)>maximum_value*percentile)[0]
    # print(prominent_idc)
    neighbor_list = [np.where(adjacency[idx, :])[0] for idx in prominent_idc]
    neighbor_cat = np.unique(np.concatenate([np.where(adjacency[idx, :]==1)[0] for idx in prominent_idc]))
    non_neighbor_cat = np.unique(np.concatenate([np.where(adjacency[idx, :]==0)[0] for idx in prominent_idc]))

    for i, idx in enumerate(neighbor_cat):
        neighbor_idc = list(np.where(adjacency[idx, :])[0])
        neighbor_idc.remove(idx)
        neighbor_idc = np.array(neighbor_idc)
        D_FOCUSS_t_smoothed[idx] = (D_FOCUSS_t[idx] + D_FOCUSS_t[neighbor_idc].sum()) / (len(neighbor_idc)+1)

    remove_idc = list(np.arange(n_dipoles))
    for neigh in neighbor_cat:
        remove_idc.remove(neigh)
    D_FOCUSS_t_smoothed[remove_idc] = 0
    sparsity = len(prominent_idc)
    return D_FOCUSS_t_smoothed, sparsity


import numpy as np
import mne
from scipy.sparse.csgraph import laplacian
from scipy.sparse.csgraph import laplacian

from .base import BaseSolver, InverseOperator

class SolverSMAP(BaseSolver):
    ''' Class for the Quadratic regularization and spatial regularization
        (S-MAP) inverse solution [1].
    
    Attributes
    ----------

    References
    ----------
    [1] Baillet, S., & Garnero, L. (1997). A Bayesian approach to introducing
    anatomo-functional priors in the EEG/MEG inverse problem. IEEE transactions
    on Biomedical Engineering, 44(5), 374-385.
    
    '''
    def __init__(self, name="S-MAP", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        LTL = self.leadfield.T @ self.leadfield 
        # n_chans, n_dipoles = self.leadfield.shape

        adjacency = mne.spatial_src_adjacency(self.forward['src'], verbose=0)
        gradient = laplacian(adjacency)
        GTG = gradient.T @ gradient
                
        inverse_operators = []
        # GG_inv = np.linalg.inv(GTG)
        for alpha in self.alphas:
            inverse_operator = np.linalg.inv(LTL + alpha * GTG) @ self.leadfield.T
            # inverse_operator = GG_inv @ self.leadfield.T @ np.linalg.inv(self.leadfield @ GG_inv @ self.leadfield.T + alpha * np.identity(n_chans))
            inverse_operators.append(inverse_operator)
        
        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        
        return self
    
